# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2019  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import xml.etree.ElementTree as ET
from uncertainties import ufloat

import hdtv.ui
from hdtv.util import Position, LockViewport
from hdtv.fitter import Fitter
from hdtv.fit import Fit
from hdtv.cmdline import HDTVCommandError

# Increase the version number if you changed something related to the xml output.
# If your change affect the reading, you should increase the major number
# and supply an appropriate new ReadFitlist function for the new xml version.
# If it is just a small change or bug fix, change the minor number.
#
# Do not remove old ReadFunctions!

VERSION = "1.5"


class FitXml(object):
    """
    Class to save and read fit lists to and from xml file
    """

    def __init__(self, spectra):
        self.spectra = spectra
        self.version = VERSION
        # Please point the following functions to the appropriate functions
        self.RestoreFromXml = self.RestoreFromXml_v1_5
        self.Xml2Fit = self.Xml2Fit_v1

    #### creating of xml #####################################################
    def WriteFitlist(self, file_object, sid=None):
        """
        Write Fitlist to file
        """
        if sid is None:
            sid = self.spectra.activeID
        try:
            fits = self.spectra.dict[sid].dict
        except KeyError:
            raise HDTVCommandError("No spectrum with id %s loaded." % sid)
        root = self.CreateXml(fits)
        # save to file
        tree = ET.ElementTree(root)
        tree.write(file_object)

    def CreateXml(self, fits):
        """
        Creates a xml tree for fits
        """
        # create xml tree
        root = ET.Element("hdtv")
        root.set("version", VERSION)
        for fit in sorted(fits.values(), key=lambda fit: fit.ID):
            root.append(self.Fit2Xml(fit))
        self._indent(root)
        return root

    def Fit2Xml(self, fit):
        """
        Creates xml element for a fit
        """
        # <fit>
        fitElement = ET.Element("fit")
        fitElement.set("peakModel", fit.fitter.peakModel.name)
        fitElement.set("integrate", str(fit.fitter.peakModel.fOptStatus["integrate"]))
        fitElement.set("likelihood", fit.fitter.peakModel.fOptStatus["likelihood"])
        fitElement.set("nParams", str(fit.fitter.backgroundModel.fParStatus["nparams"]))
        fitElement.set("chi", str(fit.chi))
        # <spectrum>
        spec = fit.spec
        specElement = ET.SubElement(fitElement, "spectrum")
        specElement.set("name", str(spec.name))
        polynom = " ".join([f"{c:e}" for c in spec.cal.GetCoeffs()])
        specElement.set("calibration", polynom)
        # <bgMarker>
        for marker in fit.bgMarkers:
            markerElement = ET.SubElement(fitElement, "bgMarker")
            # <begin>
            beginElement = ET.SubElement(markerElement, "begin")
            # <cal>
            calElement = ET.SubElement(beginElement, "cal")
            calElement.text = str(marker.p1.pos_cal)
            # <uncal>
            uncalElement = ET.SubElement(beginElement, "uncal")
            uncalElement.text = str(marker.p1.pos_uncal)
            # <end>
            endElement = ET.SubElement(markerElement, "end")
            # <cal>
            calElement = ET.SubElement(endElement, "cal")
            calElement.text = str(marker.p2.pos_cal)
            # <uncal>
            uncalElement = ET.SubElement(endElement, "uncal")
            uncalElement.text = str(marker.p2.pos_uncal)
        # <regionMarker>
        for marker in fit.regionMarkers:
            markerElement = ET.SubElement(fitElement, "regionMarker")
            # <begin>
            beginElement = ET.SubElement(markerElement, "begin")
            # <cal>
            calElement = ET.SubElement(beginElement, "cal")
            calElement.text = str(marker.p1.pos_cal)
            # <uncal>
            uncalElement = ET.SubElement(beginElement, "uncal")
            uncalElement.text = str(marker.p1.pos_uncal)
            # <p2>
            endElement = ET.SubElement(markerElement, "end")
            # <cal>
            calElement = ET.SubElement(endElement, "cal")
            calElement.text = str(marker.p2.pos_cal)
            # <uncal>
            uncalElement = ET.SubElement(endElement, "uncal")
            uncalElement.text = str(marker.p2.pos_uncal)
        # <peakMarker>
        for marker in fit.peakMarkers:
            markerElement = ET.SubElement(fitElement, "peakMarker")
            # <begin>
            positionElement = ET.SubElement(markerElement, "position")
            # <cal>
            calElement = ET.SubElement(positionElement, "cal")
            calElement.text = str(marker.p1.pos_cal)
            # <uncal>
            uncalElement = ET.SubElement(positionElement, "uncal")
            uncalElement.text = str(marker.p1.pos_uncal)
        # <background>
        bgElement = ET.SubElement(fitElement, "background")
        nparams = len(fit.bgParams)
        bgElement.set("nparams", str(nparams))
        bgElement.set("chisquare", str(fit.bgChi))
        bgElement.set("backgroundModel", fit.fitter.backgroundModel.name)
        # <coeff>
        for i in range(0, nparams):
            paramElement = ET.SubElement(bgElement, "param")
            paramElement.set("npar", str(i))
            # <value>
            valueElement = ET.SubElement(paramElement, "value")
            valueElement.text = str(fit.bgParams[i].nominal_value)
            # <error>
            errorElement = ET.SubElement(paramElement, "error")
            errorElement.text = str(fit.bgParams[i].std_dev)
        # <peak>
        for peak in fit.peaks:
            peakElement = ET.SubElement(fitElement, "peak")
            # <uncal>
            uncalElement = ET.SubElement(peakElement, "uncal")
            # Parameter
            for param in sorted(fit.fitter.fParStatus.keys()):
                paramElement = ET.SubElement(uncalElement, param)
                status = fit.fitter.fParStatus[param]
                if isinstance(status, list):
                    index = fit.peaks.index(peak)
                    status = status[index]
                paramElement.set("status", str(status))
                param = getattr(peak, param)
                if param is not None:
                    # <value>
                    if param.nominal_value is not None:
                        valueElement = ET.SubElement(paramElement, "value")
                        valueElement.text = str(param.nominal_value)
                    # <error>
                    if param.std_dev is not None:
                        errorElement = ET.SubElement(paramElement, "error")
                        errorElement.text = str(param.std_dev)
            # <cal>
            calElement = ET.SubElement(peakElement, "cal")
            # Parameter
            for param in sorted(fit.fitter.fParStatus.keys()):
                paramElement = ET.SubElement(calElement, param)
                status = fit.fitter.fParStatus[param]
                if isinstance(status, list):
                    index = fit.peaks.index(peak)
                    status = status[index]
                paramElement.set("status", str(status))
                param = getattr(peak, "%s_cal" % param)
                if param is not None:
                    # <value>
                    if param.nominal_value is not None:
                        valueElement = ET.SubElement(paramElement, "value")
                        valueElement.text = str(param.nominal_value)
                    # <error>
                    if param.std_dev is not None:
                        errorElement = ET.SubElement(paramElement, "error")
                        errorElement.text = str(param.std_dev)
            # <extras>
            extraElement = ET.SubElement(peakElement, "extras")
            for param in sorted(list(peak.extras.keys())):
                paramElement = ET.SubElement(extraElement, param)
                param = peak.extras[param]
                try:
                    if param.nominal_value is not None:
                        valueElement = ET.SubElement(paramElement, "value")
                        valueElement.text = str(param.nominal_value)
                    if param.std_dev is not None:
                        errorElement = ET.SubElement(paramElement, "error")
                        errorElement.text = str(param.std_dev)
                except BaseException:
                    paramElement.text = str(param)
        if fit.integral:
            for integral_type, integral in fit.integral.items():
                if integral is None:
                    continue
                integralElement = ET.SubElement(fitElement, "integral")
                integralElement.set("integraltype", integral_type)
                for cal_type, cal_integral in integral.items():
                    calElement = ET.SubElement(integralElement, cal_type)
                    for name, param in cal_integral.items():
                        if not name in ["id", "stat", "type"]:
                            paramElement = ET.SubElement(calElement, name)
                            if param.nominal_value is not None:
                                valueElement = ET.SubElement(paramElement, "value")
                                valueElement.text = str(param.nominal_value)
                            if param.std_dev is not None:
                                errorElement = ET.SubElement(paramElement, "error")
                                errorElement.text = str(param.std_dev)
        return fitElement

    def _indent(self, elem, level=0):
        """
        This function formats the xml in-place for prettyprinting

        Source: http://effbot.org/zone/element-lib.htm#prettyprint
        """
        i = "\n" + level * "  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for elem in elem:
                self._indent(elem, level + 1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i

    ##### Reading of xml #####################################################

    def _getPosFromElement(self, markerElement, fit=None):
        """
        Read position in energy domain from XML element.
        """
        try:
            uncal = float(markerElement.find("uncal").text)
            pos = Position(uncal, fixedInCal=False, cal=fit.cal)
        except AttributeError:
            # Try to read "cal" element if "uncal" element does not exist
            cal = float(markerElement.find("cal").text)
            pos = Position(cal, fixedInCal=True, cal=fit.cal)
        return pos

    def _readParamElement(self, paramElement):
        """
        Reads parameter info from a xml element and creates an ufloat
        """
        # status
        status = paramElement.get("status", "free")
        if status == "none":
            return None

        if status in ["free", "equal", "calculated"]:
            free = True
        else:
            free = False
        # <value>
        valueElement = paramElement.find("value")
        if valueElement is None or valueElement.text == "None":
            return None
        else:
            value = float(valueElement.text)
        # <error>
        errorElement = paramElement.find("error")
        error = float(errorElement.text)
        return ufloat(value, error, tag=free)

    def ReadFitlist(
        self,
        file_object,
        sid=None,
        calibrate=False,
        refit=False,
        interactive=True,
        fname=None,
    ):
        """
        Reads fitlist from xml files
        """

        with LockViewport(self.spectra.viewport):
            if sid is None:
                sid = self.spectra.activeID
            if sid not in self.spectra.ids:
                raise HDTVCommandError("No spectrum with id %s loaded." % sid)
            count = 0
            try:
                fname = "'{}'".format(fname or file_object.name)
            except AttributeError:
                fname = "fitlist"
            try:
                tree = ET.parse(file_object)
                root = tree.getroot()
                if not root.tag == "hdtv" or root.get("version") is None:
                    e = "this is not a valid hdtv file"
                    raise SyntaxError(e)
                # current version
                if root.get("version") == self.version:
                    count, fits = self.RestoreFromXml(
                        root,
                        sid,
                        calibrate=calibrate,
                        refit=refit,
                        interactive=interactive,
                    )
                else:
                    # old versions
                    oldversion = root.get("version")
                    hdtv.ui.warning(
                        "The XML version of this file (%s) is outdated." % oldversion
                    )
                    if oldversion == "1.4":
                        hdtv.ui.msg(
                            "But this version should be fully compatible with the new version."
                        )
                        count, fits = self.RestoreFromXml_v1_4(
                            root, sid, calibrate=calibrate, refit=refit
                        )
                    if oldversion == "1.3":
                        hdtv.ui.msg(
                            "But this version should be fully compatible with the new version."
                        )
                        count, fits = self.RestoreFromXml_v1_3(
                            root, sid, calibrate=calibrate, refit=refit
                        )
                    if oldversion == "1.2":
                        hdtv.ui.msg(
                            "But this version should be fully compatible with the new version."
                        )
                        count, fits = self.RestoreFromXml_v1_2(
                            root, sid, calibrate=calibrate, refit=refit
                        )
                    if oldversion == "1.1":
                        hdtv.ui.msg(
                            "But this version should be fully compatible with the new version."
                        )
                        count, fits = self.RestoreFromXml_v1_1(
                            root, sid, calibrate=calibrate, refit=refit
                        )
                    if oldversion == "1.0":
                        hdtv.ui.msg(
                            "Restoring only fits belonging to spectrum %s" % sid
                        )
                        hdtv.ui.msg(
                            "There may be fits belonging to other spectra in this file."
                        )
                        if interactive:
                            input("Please press enter to continue...\n")
                        count, fits = self.RestoreFromXml_v1_0(
                            root, [sid], calibrate=False, refit=refit
                        )
                    if oldversion.startswith("0"):
                        hdtv.ui.msg(
                            "Only the fit markers have been saved in this file."
                        )
                        hdtv.ui.msg("All the fits therefore have to be repeated.")
                        hdtv.ui.msg("This will take some time...")
                        if interactive:
                            input("Please press enter to continue...\n")
                        count, fits = self.RestoreFromXml_v0(root, True)
            except SyntaxError as e:
                raise HDTVCommandError(f"Error reading fits from {fname}.")
            else:
                msg = "%s loaded: " % (fname)
                if count == 1:
                    msg += "1 fit restored."
                else:
                    msg += "%d fits restored." % count
                hdtv.ui.msg(msg)
                return count, fits

    #### version 1* ###############################################################
    def RestoreFromXml_v1_5(
        self, root, sid, calibrate=False, refit=False, interactive=True
    ):
        """
        Restores fits from xml file (version = 1.5)

        Changes to version 1.4:
        Notation of the parameters for the background model has been modified to
        allow for a more general interpretation as arbitrary fit parameters.
        No big change!
        """
        return self.RestoreFromXml_v1_4(
            root, sid, calibrate, refit, interactive=interactive
        )

    def RestoreFromXml_v1_4(
        self, root, sid, calibrate=False, refit=False, interactive=True
    ):
        """
        Restores fits from xml file (version = 1.4)

        Changes to version 1.3:
        Information about the integral over the fit region (and bg, if
        available) are supplied. No big change!
        """

        spec = self.spectra.dict[sid]
        count = 0
        do_fit = ""
        fits = list()
        spec_name_last = ""
        for fitElement in root.findall("fit"):
            if calibrate:
                spectrum = fitElement.find("spectrum")
                spec_name = spectrum.get("name")
                spec_cal = spectrum.get("calibration")
                if spec_cal and spec_name != spec_name_last and spec_name:
                    if spec_name != spec.name:
                        hdtv.ui.warning(
                            f"Applying calibration for '{spec_name}' to '{spec.name}'."
                        )
                    spec_name_last = spec_name
                    cal = [float(c) for c in spec_cal.split()]
                    self.spectra.ApplyCalibration([sid], cal)
                    hdtv.ui.debug(f"Applying calibration {spec_cal}.")

            (fit, success) = self.Xml2Fit_v1(fitElement, calibration=spec.cal)
            # restore fit
            if success and not refit:
                try:
                    fit.Restore(spec=spec)
                except (TypeError, IndexError) as err:
                    hdtv.ui.debug("An exception occurred while restoring the peaks")
                    hdtv.ui.debug(err)
                    success = False
            # deal with failure
            if not success or refit:
                if refit:
                    do_fit = "a"
                if do_fit not in ["V", "v"]:  # Ne(v)er
                    if do_fit not in ["A", "a"]:  # (A)lways
                        do_fit = None
                    while do_fit not in ["Y", "y", "N", "n", "", "A", "a", "V", "v"]:
                        question = "Could not restore fit. Refit? [(Y)es/(n)o/(a)lways/ne(v)er]"
                        do_fit = input(question)
                    if do_fit in ["Y", "y", "", "A", "a"]:
                        fit.FitPeakFunc(spec)
            # finish this fit
            fits.append(fit)
        # add fits to spectrum
        for fit in fits:
            ID = spec.Insert(fit)
            count += 1
            if sid not in self.spectra.visible:
                fit.Hide()
        return count, fits

    def RestoreFromXml_v1_3(
        self, root, sid, calibrate=False, refit=False, interactive=True
    ):
        """
        Restores fits from xml file (version = 1.3)

        Changes to version 1.2:
        Now it is possible to store additional user supplied parameter for
        each peak. No big change!
        """
        return self.RestoreFromXml_v1_4(
            root, sid, calibrate, refit, interactive=interactive
        )

    def RestoreFromXml_v1_2(
        self, root, sid, calibrate=False, refit=False, interactive=True
    ):
        """
        Restores fits from xml file (version = 1.2)

        Changes to version 1.1:
        Calibrated AND Uncalibrated marker positions are saved in version 1.2,
        where as in version 1.1 only one of both has been saved. No big change!
        """
        return self.RestoreFromXml_v1_3(
            root, sid, calibrate, refit, interactive=interactive
        )

    def RestoreFromXml_v1_1(
        self, root, sid, calibrate=False, refit=False, interactive=True
    ):
        """
        Restores fits from xml file (version = 1.1)

        Changes compared to version 1.0:
        In this version a file contains only fits belonging to one spectra
        this changes the xml hierachy a little, the fit elements are now direct
        childs of the root element. The spectrum element is still there, but now
        a child of each fit element and not longer used in hdtv for anything
        """
        return self.RestoreFromXml_v1_2(
            root, sid, calibrate, refit, interactive=interactive
        )

    def RestoreFromXml_v1_0(
        self, root, sids=None, calibrate=False, refit=False, interactive=False
    ):
        """
        Restores fits from xml file (version = 1.0)

        This version is able to restore the fits from the XML.
        Changes compared to version 0.1:
            * changed some node namings:
                ** <background> -> <bgMarker>
                ** <region>     -> <regionMarker>
                ** <peak>       -> <peakMarker>
                ** <result>     -> <peak>
            * new node for the background polynomial <background>
            * saved chisquare of the fits
            * calibrated and uncalibrated values for each peak parameter
        """
        do_fit = None
        count = 0
        # default is to look for all loaded spectra
        if sids is None:
            sids = self.spectra.ids
        # create an index of spectra that are saved in xml
        index = dict()
        for specElement in root.iter():
            name = specElement.get("name")
            index[name] = specElement
        # <spectrum>
        all_fits = list()
        for sid in sids:
            spec = self.spectra.dict[sid]
            try:
                specElement = index[spec.name]
            except KeyError:
                #    msg ="No informations for spectrum %s (id=%d) in file" %(spec, sid)
                #    hdtv.ui.warning(msg)
                continue
            # load the calibration from file
            if calibrate:
                try:
                    calibration = list(
                        map(float, specElement.get("calibration").split())
                    )
                    spec.cal = calibration
                except AttributeError:
                    # No calibration was saved
                    msg = "Could not read calibration for spectrum %s (id=%d)" % (
                        spec,
                        sid,
                    )
                    hdtv.ui.warning(msg)
            # <fits>
            fits = list()
            for fitElement in specElement.findall("fit"):
                (fit, success) = self.Xml2Fit_v1(fitElement, spec.cal)
                count = count + 1
                # restore fit
                if success and not refit:
                    try:
                        fit.Restore(spec=spec)
                    except (TypeError, IndexError):
                        success = False
                # deal with failure
                if not success or refit:
                    if refit:
                        do_fit = "a"
                    if do_fit not in ["V", "v"]:  # Ne(v)er
                        if do_fit not in ["A", "a"]:  # (A)lways
                            do_fit = None
                        while do_fit not in [
                            "Y",
                            "y",
                            "N",
                            "n",
                            "",
                            "A",
                            "a",
                            "V",
                            "v",
                        ]:
                            question = "Could not restore fit. Refit? [(Y)es/(n)o/(a)lways/ne(v)er]"
                            do_fit = input(question)
                        if do_fit in ["Y", "y", "", "A", "a"]:
                            fit.FitPeakFunc(spec)
                # finish this fit
                ID = spec.Insert(fit)
                if sid not in self.spectra.visible:
                    fit.Hide()
                fits.append(fit)
                all_fits.append(fit)
        return count, all_fits

    def Xml2Fit_v1(self, fitElement, calibration=None):
        """
        Creates a fit object from information found in a xml file
        """
        # <fit>
        success = True
        peakModel = fitElement.get("peakModel")
        bgElement = fitElement.find("background")
        # Simple fix for older xml file versions, where the only background
        # model was a polynomial, and therefore it did not have to be stored

        try:
            backgroundModel = bgElement.get("backgroundModel")
        except AttributeError:
            backgroundModel = "polynomial"
        if backgroundModel is None:
            backgroundModel = "polynomial"
        fitter = Fitter(peakModel, backgroundModel)
        fit = Fit(fitter, cal=calibration)

        for parname in ["integrate", "likelihood"]:
            if parname in fitElement.attrib:
                fitter.SetParameter(parname, fitElement.get(parname))
        try:
            fit.chi = float(fitElement.get("chi"))
        except ValueError:
            fit.chi = None
        # <bgMarker>
        for bgElement in fitElement.findall("bgMarker"):
            # Read begin marker
            beginElement = bgElement.find("begin")
            begin = self._getPosFromElement(beginElement, fit)
            fit.ChangeMarker("bg", begin, "set")
            # Read end marker
            endElement = bgElement.find("end")
            end = self._getPosFromElement(endElement, fit)
            fit.ChangeMarker("bg", end, "set")
        # <regionMarker>
        for regionElement in fitElement.findall("regionMarker"):
            # Read begin marker
            beginElement = regionElement.find("begin")
            begin = self._getPosFromElement(beginElement, fit)
            fit.ChangeMarker("region", begin, "set")
            # Read end marker
            endElement = regionElement.find("end")
            end = self._getPosFromElement(endElement, fit)
            fit.ChangeMarker("region", end, "set")
            # <peakMarker>
        for peakElement in fitElement.findall("peakMarker"):
            # Read position marker
            posElement = peakElement.find("position")
            pos = self._getPosFromElement(posElement, fit)
            fit.ChangeMarker("peak", pos, "set")
        # <background>
        bgElement = fitElement.find("background")
        if bgElement:
            try:
                fit.bgChi = float(bgElement.get("chisquare"))
            except ValueError:
                pass
            params = list()

            # Distinguish between old notation of background parameters (coeff, ncoeff),
            # which interprets the parameters as coefficients of a polynomial, and the
            # new notation (params, npar), which interprets them as arbitrary parameters.
            paramCounterName = "npar"
            paramElements = bgElement.findall("param")
            if not paramElements:
                paramElements = bgElement.findall("coeff")
                paramCounterName = "deg"

            for paramElement in paramElements:
                npar = int(paramElement.get(paramCounterName))
                # <value>
                valueElement = paramElement.find("value")
                value = float(valueElement.text)
                # <error>
                errorElement = paramElement.find("error")
                error = float(errorElement.text)
                param = ufloat(value, error)
                params.append([npar, param])
            params.sort()
            fit.bgParams = [p[1] for p in params]

        # <peak>
        statusdict = dict()
        for peakElement in fitElement.findall("peak"):
            # <uncal>
            uncalElement = peakElement.find("uncal")
            parameter = dict()
            for paramElement in uncalElement:
                # parameter value/error
                name = paramElement.tag
                parameter[name] = self._readParamElement(paramElement)
                # parameter status
                status = paramElement.get("status", "free")
                try:
                    statusdict[name].append(status)
                except KeyError:
                    statusdict[name] = [status]
            # <extras>
            extraElement = peakElement.find("extras")
            extras = dict()
            if extraElement is not None:
                for paramElement in extraElement:
                    name = paramElement.tag
                    if len(paramElement) == 1:
                        extras[name] = paramElement.text
                    else:
                        # <value>
                        valueElement = paramElement.find("value")
                        value = float(valueElement.text)
                        # <error>
                        errorElement = paramElement.find("error")
                        error = float(errorElement.text)
                        extras[name] = ufloat(value, error)
            # create peak
            try:
                peak = fit.fitter.peakModel.Peak(cal=calibration, **parameter)
            except TypeError:
                hdtv.ui.error("Error reading peak with parameters: %s" % str(parameter))
                success = False
                continue
            peak.extras = extras
            fit.peaks.append(peak)
            # set parameter status of fitter
            for name in list(statusdict.keys()):
                # check if status is the same for all peaks
                check = set(statusdict[name])
                if len(check) == 1:
                    status = statusdict[name][0]
                else:
                    status = statusdict[name]
                fitter.SetParameter(name, status)
        integrals = dict()
        for integral in fitElement.findall("integral"):
            integral_type = integral.get("integraltype")
            integrals[integral_type] = dict()
            for calElement in integral:
                cal_type = calElement.tag
                integrals[integral_type][cal_type] = dict()
                for paramElement in calElement:
                    # <value>
                    valueElement = paramElement.find("value")
                    value = float(valueElement.text)
                    # <error>
                    errorElement = paramElement.find("error")
                    error = float(errorElement.text)
                    coeff = ufloat(value, error)
                    integrals[integral_type][cal_type][paramElement.tag] = coeff
        if not integrals:
            integrals = None
        for integral_type in ["sub", "bg"]:
            if integrals and not integral_type in integrals:
                integrals[integral_type] = None
        fit.integral = integrals
        return (fit, success)

    ##### version 0.* ########################################################

    def RestoreFromXml_v0(self, root, do_fit=False):
        """
        Reads fits from xml file (version = 0.*)

        Note: For performance reasons this does not reconstruct the fit results.
        It only sets the markers and restores the status of the fitter. The user
        must repeat the fit, if he/she wants to see the results again.
        (This should be improved in later versions.)
        """
        count = 0
        spectra = self.spectra
        # <spectrum>
        for specElement in root.iter():
            name = specElement.get("name")
            # find this spectrum from Element in the real world
            spec = None
            for sid in spectra.ids:
                if spectra.dict[sid].name == name:
                    spec = spectra.dict[sid]
            # maybe the spectrum that is referred to in XML is currently not
            # loaded
            if spec is None:
                continue
            # <fit>
            for fitElement in specElement:
                count = count + 1
                peakModel = fitElement.get("peakModel")
                # Simple fix for older xml file versions, where the only background
                # model was a polynomial, and therefore it did not have to be stored
                fitter = Fitter(peakModel, "polynomial")
                # <result>
                params = dict()
                for resultElement in fitElement.findall("result"):
                    for paramElement in resultElement:
                        parname = paramElement.tag
                        status = paramElement.get("status")
                        try:
                            params[parname].append(status)
                        except KeyError:
                            params[parname] = [status]
                for parname in list(params.keys()):
                    status = params[parname]
                    fitter.SetParameter(parname, status)
                fit = Fit(fitter, cal=spec.cal)
                # <background>
                for bgElement in fitElement.findall("background"):
                    # Read begin/p1 marker
                    beginElement = bgElement.find("begin")
                    if beginElement is None:  # Maybe old Element (ver 0.1)
                        beginElement = bgElement.find("p1")
                    begin = float(beginElement.find("uncal").text)
                    fit.ChangeMarker("bg", fit.cal.Ch2E(begin), "set")
                    # Read end/p2 marker
                    endElement = bgElement.find("end")
                    if endElement is None:  # Maybe old Element (ver 0.1)
                        endElement = bgElement.find("p2")
                    end = float(endElement.find("uncal").text)
                    fit.ChangeMarker("bg", fit.cal.Ch2E(end), "set")
                # <region>
                for regionElement in fitElement.findall("region"):
                    # Read begin/p1 marker
                    beginElement = regionElement.find("begin")
                    if beginElement is None:  # Maybe old Element (ver 0.1)
                        beginElement = regionElement.find("p1")
                    begin = float(beginElement.find("uncal").text)
                    fit.ChangeMarker("region", fit.cal.Ch2E(begin), "set")
                    # Read end/p2 marker
                    endElement = regionElement.find("end")
                    if endElement is None:  # Maybe old Element (ver 0.1)
                        endElement = regionElement.find("p2")
                    end = float(endElement.find("uncal").text)
                    fit.ChangeMarker("region", fit.cal.Ch2E(end), "set")
                # <peak>
                for peakElement in fitElement.findall("peak"):
                    # Read position/p1 marker
                    posElement = peakElement.find("position")
                    if posElement is None:
                        posElement = peakElement.find("p1")
                    pos = float(posElement.find("uncal").text)
                    fit.ChangeMarker("peak", fit.cal.Ch2E(pos), "set")
                if do_fit:
                    fit.FitPeakFunc(spec)
                spec.Insert(fit)
            return count, [fit]



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import traceback
import ROOT
import hdtv.options
import hdtv.color
import hdtv.cmdline
import hdtv.ui
from hdtv.marker import MarkerCollection
import hdtv.rootext.display
from hdtv.cmdline import HDTVCommandAbort, HDTVCommandError


class HotkeyList(object):
    "Class to handle multi-key hotkeys."

    def __init__(self):
        self.fKeyCmds = dict()
        self.fCurNode = self.fKeyCmds

    def AddHotkey(self, key, cmd):
        """
        Adds a hotkey to the hotkey table. If key is a list,
        it is taken as a multi-key hotkey. The function refuses
        to overwrite definitions that do not match exactly, i.e.
        once you have defined hotkey x, you can no longer define
        the multi-key combination xy (since that would have to
        overwrite x). You can overwrite matching combinations,
        though (i.e. redefine x in the example above).
        """
        curNode = self.fKeyCmds
        if isinstance(key, list):
            for k in key[0:-1]:
                try:
                    curNode = curNode[k]
                    if not isinstance(curNode, dict):
                        raise RuntimeError("Refusing to overwrite non-matching hotkey")
                except KeyError:
                    curNode[k] = dict()
                    curNode = curNode[k]

            key = key[-1]

        curNode[key] = wrap_cmd(cmd)

    def HandleHotkey(self, key):
        """
        Handles a key press. Returns True if the input was a
        valid hotkey, False if it was not, and None if further
        input is needed (i.e. if the input could be part of a
        multi-key hotkey)
        """
        try:
            node = self.fCurNode[key]
        except KeyError:
            self.ResetHotkeyState()
            return False

        if isinstance(node, dict):
            self.fCurNode = node
            return None
        else:
            node()
            self.ResetHotkeyState()
            return True

    def ResetHotkeyState(self):
        """
        Reset possible waiting state for a multi-key hotkey.
        """
        self.fCurNode = self.fKeyCmds


class KeyHandler(HotkeyList):
    """
    Hotkey handler, adding support for a status line (that can be (ab)used as
    a text entry)
    """

    def __init__(self):
        super(KeyHandler, self).__init__()

        self.fEditMode = False  # Status bar currently used as text entry
        # Modifier keys
        self.MODIFIER_KEYS = (
            ROOT.kKey_Shift,
            ROOT.kKey_Control,
            ROOT.kKey_Meta,
            ROOT.kKey_Alt,
            ROOT.kKey_CapsLock,
            ROOT.kKey_NumLock,
            ROOT.kKey_ScrollLock,
        )

        # Parent class must provide self.viewer and self.viewport!

    def EnterEditMode(self, prompt, handler):
        """
        Enter edit mode (mode where the status line is used as a text entry)
        """
        self.fEditMode = True
        self.fEditStr = ""
        self.fEditPrompt = prompt
        self.fEditHandler = wrap_cmd(handler)
        self.viewport.SetStatusText(self.fEditPrompt)

    def EditKeyHandler(self):
        """
        Key handler in edit mode
        """
        try:
            keyStr = self.viewer.fKeyStr
        except UnicodeDecodeError as err:
            keyStr = err.args[1].split(b"\x00")[0].decode("utf-8")

        if self.viewer.fKeySym == ROOT.kKey_Backspace:
            self.fEditStr = self.fEditStr[0:-1]
            self.viewport.SetStatusText(self.fEditPrompt + self.fEditStr)
        elif (
            self.viewer.fKeySym == ROOT.kKey_Return
            or self.viewer.fKeySym == ROOT.kKey_Enter
        ):
            self.fEditMode = False
            self.viewport.SetStatusText("")
            self.fEditHandler(self.fEditStr)
        elif keyStr != "":
            self.fEditStr += keyStr[0]
            self.viewport.SetStatusText(self.fEditPrompt + self.fEditStr)
        else:
            return False

        return True

    def KeyHandler(self):
        """
        Key handler (handles hotkeys and calls EditKeyHandler in edit mode)
        """
        # Filter unknown and modifier keys
        if (
            self.viewer.fKeySym in self.MODIFIER_KEYS
            or self.viewer.fKeySym == ROOT.kKey_Unknown
        ):
            return

        # ESC aborts
        if self.viewer.fKeySym == ROOT.kKey_Escape:
            self.ResetHotkeyState()
            self.fEditMode = False
            self.keyString = ""
            self.viewport.SetStatusText("")
            return True

        # There are two modes: the ``edit'' mode, in which the status
        # bar is abused as a text entry, and the normal mode, in which
        # the keys act as hotkeys.
        if self.fEditMode:
            handled = self.EditKeyHandler()
        else:
            try:
                keyStr = self.viewer.fKeyStr
            except UnicodeDecodeError as err:
                keyStr = err.args[1].split(b"\x00")[0].decode("utf-8")

            if not keyStr:
                keyStr = "<?>"

            handled = self.HandleHotkey(self.viewer.fKeySym)

            if handled is None:
                self.keyString += keyStr
                self.viewport.SetStatusText("Command: %s" % self.keyString)
            elif handled == False:
                self.keyString += keyStr
                self.viewport.SetStatusText("Invalid hotkey %s" % self.keyString)
                self.keyString = ""
            else:
                if not self.fEditMode:
                    self.viewport.SetStatusText("")
                self.keyString = ""

        return handled


class Window(KeyHandler):
    """
    Base class of a window object

    This class provides basic key handling for zooming and scrolling.
    """

    def __init__(self):
        super(Window, self).__init__()

        self.viewer = ROOT.HDTV.Display.Viewer()
        self.viewport = self.viewer.GetViewport()
        self._dispatchers = list()

        # Handle closing of the main window (with an application exit)
        disp = ROOT.TPyDispatcher(hdtv.cmdline.AsyncExit)
        self.viewer.Connect("CloseWindow()", "TPyDispatcher", disp, "Dispatch()")
        self._dispatchers.append(disp)

        self.XZoomMarkers = MarkerCollection(
            "X", paired=True, maxnum=1, color=hdtv.color.zoom
        )
        self.XZoomMarkers.Draw(self.viewport)
        self.YZoomMarkers = MarkerCollection(
            "Y", paired=True, maxnum=1, color=hdtv.color.zoom
        )
        self.YZoomMarkers.Draw(self.viewport)

        # Key Handling
        disp = ROOT.TPyDispatcher(self.KeyHandler)
        self.viewer.Connect("KeyPressed()", "TPyDispatcher", disp, "Dispatch()")
        self._dispatchers.append(disp)

        self.keyString = ""
        self.AddHotkey(ROOT.kKey_u, lambda: self.viewport.Update())
        # toggle spectrum display
        self.AddHotkey(ROOT.kKey_l, self.viewport.ToggleLogScale)
        self.AddHotkey(ROOT.kKey_A, self.viewport.ToggleYAutoScale)
        self.AddHotkey(ROOT.kKey_Exclam, self.viewport.ToggleUseNorm)
        # x directions
        self.AddHotkey(ROOT.kKey_Space, self.SetXZoomMarker)
        self.AddHotkey(ROOT.kKey_x, self.ExpandX)
        self.AddHotkey(ROOT.kKey_Right, lambda: self.viewport.ShiftXOffset(0.1))
        self.AddHotkey(ROOT.kKey_Left, lambda: self.viewport.ShiftXOffset(-0.1))
        self.AddHotkey(ROOT.kKey_Greater, lambda: self.viewport.ShiftXOffset(0.1))
        self.AddHotkey(ROOT.kKey_Less, lambda: self.viewport.ShiftXOffset(-0.1))
        self.AddHotkey(ROOT.kKey_Return, lambda: self.viewport.Update(True))
        self.AddHotkey(ROOT.kKey_Enter, lambda: self.viewport.Update(True))
        self.AddHotkey(
            ROOT.kKey_Bar, lambda: self.viewport.SetXCenter(self.viewport.GetCursorX())
        )
        self.AddHotkey(ROOT.kKey_1, lambda: self.viewport.XZoomAroundCursor(2.0))
        self.AddHotkey(ROOT.kKey_0, lambda: self.viewport.XZoomAroundCursor(0.5))
        # y direction
        self.AddHotkey(ROOT.kKey_h, self.SetYZoomMarker)
        self.AddHotkey(ROOT.kKey_y, self.ExpandY)
        self.AddHotkey(ROOT.kKey_Up, lambda: self.viewport.ShiftYOffset(0.1))
        self.AddHotkey(ROOT.kKey_Down, lambda: self.viewport.ShiftYOffset(-0.1))
        self.AddHotkey(ROOT.kKey_Z, lambda: self.viewport.YZoomAroundCursor(2.0))
        self.AddHotkey(ROOT.kKey_X, lambda: self.viewport.YZoomAroundCursor(0.5))
        # expand in all directions
        self.AddHotkey(ROOT.kKey_e, self.Expand)

        self.AddHotkey(
            ROOT.kKey_i,
            lambda: self.EnterEditMode(prompt="Position: ", handler=self.GoToPosition),
        )

        # Register configuration variables
        opt = hdtv.options.Option(
            default=self.viewport.GetYMinVisibleRegion(),
            parse=lambda x: float(x),
            changeCallback=self.YMinVisibleRegionChanged,
        )
        hdtv.options.RegisterOption("display.YMinVisibleRegion", opt)

        opt = hdtv.options.Option(
            default=self.viewport.GetDarkMode(),
            parse=hdtv.options.parse_bool,
            changeCallback=self.SetDarkMode,
        )
        hdtv.options.RegisterOption("display.mode.dark", opt)

        prog = "window view center"
        description = "center window to position"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-w", "--width", type=float, help="width of window", default=100.0
        )
        parser.add_argument("position", type=float)
        hdtv.cmdline.AddCommand(prog, self.GoToPosition, parser=parser)

        prog = "window view region"
        description = "show region in window"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("start", type=float)
        parser.add_argument("end", type=float)
        hdtv.cmdline.AddCommand(prog, self.ViewRegion, parser=parser)

    def YMinVisibleRegionChanged(self, opt):
        self.viewport.SetYMinVisibleRegion(opt.Get())

    def SetDarkMode(self, opt):
        """
        Switch between dark and light mode.
        """
        self.viewport.SetDarkMode(opt.Get())
        try:
            from hdtv.plugins.matInterface import matrix_interface

            for viewport in matrix_interface.matviews:
                viewport.SetDarkMode(opt.Get())
        except:
            pass

    def ViewRegion(self, args):
        """
        Zoom and move viewport to show a region
        """
        width = abs(args.end - args.start)
        center = args.start + width / 2.0
        self.viewport.SetXVisibleRegion(width)
        self.viewport.SetXCenter(center)

    def GoToPosition(self, args):
        """
        Move viewport to be centered around a position
        """
        try:
            width = args.width
        except AttributeError:
            width = 100.0
        try:
            center = float(args)
        except TypeError:
            center = float(args.position)
        self.viewport.SetXVisibleRegion(width)
        self.viewport.SetXCenter(center)

    def FocusObjects(self, objs):
        """
        Move and stretch viewport to show multiple objs
        """
        xdimensions = ()
        # Get dimensions of objects
        for obj in objs:
            try:
                if obj.xdimensions is not None:
                    xdimensions += obj.xdimensions
            except BaseException:
                # ignore objects without xdimensions
                pass
        # calulate
        if len(xdimensions) > 0:
            view_width = max(xdimensions) - min(xdimensions)
            view_width *= 1.2
            if view_width < 50.0:
                view_width = 50.0  # TODO: make this configurable
            view_center = (max(xdimensions) + min(xdimensions)) / 2.0
            # change viewport
            self.viewport.SetXVisibleRegion(view_width)
            self.viewport.SetXCenter(view_center)

    def IsInVisibleRegion(self, obj, part=False):
        """
        Check if obj is in the visible region of the viewport
        """
        try:
            xdim = obj.xdimensions
        except BaseException:
            xdim = None
        if xdim is None:  # Object has no dimensions
            return True
        # get viewport limits
        viewport_start = self.viewport.GetXOffset()
        viewport_end = viewport_start + self.viewport.GetXVisibleRegion()
        # check if the full object is visible
        if not part:
            if (xdim[0] > viewport_start) and (xdim[1] < viewport_end):
                return True
            else:
                return False
        # check if a part of the object is visible
        if part:
            if (xdim[0] < viewport_start) and (xdim[1] < viewport_start):
                return False
            elif (xdim[0] > viewport_end) and (xdim[1] > viewport_end):
                return False
            else:
                return True

    def ExpandX(self):
        """
        expand in X direction
        """
        self._Expand("X")

    def ExpandY(self):
        """
        expand in Y direction
        """
        self._Expand("Y")

    def Expand(self):
        """
        expand in X and in Y direction
        """
        self._Expand("X")
        self._Expand("Y")

    def _Expand(self, xytype):
        """
        expand the display to show the region between the zoom markers (X or Y),
        or the full spectrum if not zoom markers are set.
        """
        # check the input
        if xytype not in ["X", "Y"]:
            hdtv.ui.error(
                "invalid parameter %s to the private function _Expand" % xytype
            )
            return

        zoomMarkers = getattr(self, "%sZoomMarkers" % xytype)
        if len(zoomMarkers) == 1:
            zm = zoomMarkers[0]
            p1 = zm.p1.pos_cal
            if zm.p2 is None:
                p2 = 0.0
            else:
                p2 = zm.p2.pos_cal
            setOffset = getattr(self.viewport, "Set%sOffset" % xytype)
            setOffset(min(p1, p2))
            setVisibleRegion = getattr(self.viewport, "Set%sVisibleRegion" % xytype)
            setVisibleRegion(abs(p2 - p1))
            getattr(self, "%sZoomMarkers" % xytype).pop()
        else:
            if xytype == "X":
                self.viewport.ShowAll()
            elif xytype == "Y":
                self.viewport.YAutoScaleOnce()

    def SetXZoomMarker(self, pos=None):
        """
        set a X zoom marker
        """
        if pos is None:
            pos = self.viewport.GetCursorX()
        self.XZoomMarkers.SetMarker(pos)

    def SetYZoomMarker(self, pos=None):
        """
        set a Y zoom marker
        """
        if pos is None:
            pos = self.viewport.GetCursorY()
        self.YZoomMarkers.SetMarker(pos)


def wrap_cmd(cmd):
    def _wrapper(*args, **kwargs):
        try:
            cmd(*args, **kwargs)
        except hdtv.cmdline.HDTVCommandAbort as msg:
            if msg.value:
                hdtv.ui.error(msg.value)
        except (hdtv.cmdline.HDTVCommandError, Exception) as msg:
            try:
                if msg.value:
                    hdtv.ui.error(msg.value)
            except AttributeError:
                hdtv.ui.error(str(msg))
            hdtv.ui.debug(traceback.format_exc())

    return _wrapper



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
This script tests the interplay of fitting and restoring fits with calibration
"""

import os

import pytest

from numpy import linspace, ones, savetxt, sqrt
from scipy.stats import norm

from tests.helpers.utils import setup_io, redirect_stdout, isclose
from tests.helpers.fixtures import temp_file

from hdtv.util import monkey_patch_ui

monkey_patch_ui()

import __main__

import hdtv.session

try:
    __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass
spectra = __main__.spectra

from hdtv.plugins.specInterface import spec_interface
from hdtv.plugins.fitInterface import fit_interface
from hdtv.plugins.fitlist import fitxml

testspectrum = os.path.join(os.path.curdir, "tests", "share", "osiris_bg.spc")
testspectrum_h1 = os.path.join(os.path.curdir, "tests", "share", "binning_h1.tv")
testspectrum_h2 = os.path.join(os.path.curdir, "tests", "share", "binning_h2.tv")


NBINS = 2000
BG_PER_BIN = 0.0
PEAK_VOLUME = 1e6
PEAK_WIDTH = 15.0


@pytest.fixture(autouse=True)
def prepare():
    fit_interface.ResetFitterParameters()
    hdtv.options.Set("table", "classic")
    hdtv.options.Set("uncertainties", "short")
    spec_interface.LoadSpectra(testspectrum)
    yield
    spectra.Clear()

    bins_h1 = linspace(0.5, NBINS + 0.5, NBINS)
    spectrum_h1 = BG_PER_BIN * ones(NBINS)
    spectrum_h1 = spectrum_h1 + PEAK_VOLUME * norm.pdf(
        bins_h1, loc=0.5 * NBINS, scale=PEAK_WIDTH
    )
    savetxt(testspectrum_h1, spectrum_h1)

    bins_h2 = linspace(0.5, NBINS + 0.5, int(0.5 * NBINS))
    spectrum_h2 = 0.5 * BG_PER_BIN * ones(int(0.5 * NBINS))
    spectrum_h2 = spectrum_h2 + 2.0 * PEAK_VOLUME * norm.pdf(
        bins_h2, loc=0.5 * NBINS, scale=PEAK_WIDTH
    )
    savetxt(testspectrum_h2, spectrum_h2)


def get_markers(fit=None):
    if fit is None:
        fit = spectra.workFit
    return [
        fit.regionMarkers[0].p1.pos_uncal,
        fit.regionMarkers[0].p1.pos_cal,
        fit.regionMarkers[0].p2.pos_uncal,
        fit.regionMarkers[0].p2.pos_cal,
        fit.peakMarkers[0].p1.pos_uncal,
        fit.peakMarkers[0].p1.pos_cal,
    ]


def markers_consistent(calfactor=1.0, region1=None, region2=None, fit=None):
    markers = get_markers(fit)
    assert isclose(markers[0], markers[1] / calfactor)
    assert isclose(markers[2], markers[3] / calfactor)
    assert isclose(markers[4], markers[5] / calfactor)
    if region1:
        assert isclose(markers[1], region1)
    if region2:
        assert isclose(markers[3], region2)


def list_fit():
    f, ferr = setup_io(2)
    with redirect_stdout(f, ferr):
        fit_interface.ListFits()
        fit_interface.ListIntegrals()
    assert ferr.getvalue().strip() == ""
    return f.getvalue().strip()


@pytest.mark.parametrize("region1, region2, peak", [(1450.0, 1470.0, 1460.0)])
def test_mark(region1, region2, peak):
    """
    Set Marker and calibrate afterwards
    Marker will be kept fixed, but the spectrum changes
    """
    spectra.SetMarker("region", region1)
    spectra.SetMarker("region", region2)
    spectra.SetMarker("peak", peak)
    markers_before = get_markers()
    assert markers_before == [region1, region1, region2, region2, peak, peak]
    return markers_before


@pytest.mark.parametrize("region1, region2, peak", [(1450.0, 1470.0, 1460.0)])
def test_mark_and_calibrate(region1, region2, peak):
    """
    Set Marker and calibrate afterwards
    Marker will be kept fixed, but the spectrum changes
    """
    markers_before = test_mark(region1, region2, peak)
    spectra.ApplyCalibration("0", [0, 2])
    markers_after = get_markers()
    assert all([x == y for (x, y) in zip(markers_before, markers_after)])


@pytest.mark.parametrize("region1, region2, peak", [(2900.0, 2940.0, 2920.0)])
def test_marker_in_calibrated_spectrum(region1, region2, peak):
    """
    Set Marker in calibrated spectrum
    Marker should be appear at the right position relative to spectrum
    """
    spectra.ApplyCalibration("0", [0, 2])
    test_mark(region1, region2, peak)


@pytest.mark.parametrize("calfactor", [2.0])
@pytest.mark.parametrize("region1, region2, peak", [(2900.0, 2940.0, 2920.0)])
def test_fit_in_calibrated_spectrum(region1, region2, peak, calfactor):
    """
    Fit in calibrated spectrum
    Fit should be fixed relative to spectrum
    """
    spectra.ApplyCalibration("0", [0, calfactor])
    spectra.SetMarker("region", region1)
    spectra.SetMarker("region", region2)
    spectra.SetMarker("peak", peak)
    spectra.ExecuteFit()
    markers_consistent(calfactor, region1, region2)


# Check that bin sizes are handled in fitted peak volume (see also test_root_fit_volume_with_binning)
@pytest.mark.parametrize(
    "calfactor, region1, region2, peak, expected_volume",
    [
        (1.0, 500.0, 1500.0, 1000.0, PEAK_VOLUME),
        (2.0, 500.0, 1500.0, 1000.0, PEAK_VOLUME),
    ],
)
def test_mfile_fit_volume_with_binning(
    region1, region2, peak, expected_volume, calfactor, temp_file
):
    spectra.Clear()

    bins = linspace(0.5, NBINS + 0.5, int(NBINS / calfactor))
    spectrum = 1.0 / calfactor * BG_PER_BIN * ones(int(NBINS / calfactor))
    spectrum = spectrum + calfactor * PEAK_VOLUME * norm.pdf(
        bins, loc=0.5 * NBINS, scale=PEAK_WIDTH
    )
    savetxt(temp_file, spectrum)

    spec_interface.LoadSpectra(temp_file)
    spectra.ApplyCalibration("0", [0, calfactor])

    spectra.SetMarker("region", region1)
    spectra.SetMarker("region", region2)
    spectra.SetMarker("peak", peak)
    spectra.ExecuteFit()
    fit = spectra.workFit
    fit_volume = fit.ExtractParams()[0][0]["vol"].nominal_value
    integral_volume = fit.ExtractIntegralParams()[0][0]["vol"].nominal_value

    assert isclose(integral_volume, expected_volume, abs_tol=sqrt(PEAK_VOLUME))
    assert isclose(fit_volume, expected_volume, abs_tol=sqrt(PEAK_VOLUME))


@pytest.mark.parametrize("calfactor1", [2.1, 1.9])
@pytest.mark.parametrize("calfactor2", [0.4, 0.6])
@pytest.mark.parametrize("region1, region2, peak", [(2900.0, 2940.0, 2920.0)])
def test_change_calibrated_fit(region1, region2, peak, calfactor1, calfactor2):
    """
    Change calibration after fit is executed
    Fit should move with spectrum to new position
    """
    test_fit_in_calibrated_spectrum(region1, region2, peak, calfactor1)
    spectra.ApplyCalibration("0", [0, calfactor2])
    markers_consistent(
        calfactor2, region1 / calfactor1 * calfactor2, region2 / calfactor1 * calfactor2
    )


@pytest.mark.parametrize("calfactor1", [2.1, 1.9])
@pytest.mark.parametrize("calfactor2", [0.4, 0.6])
@pytest.mark.parametrize("region1, region2, peak", [(2900.0, 2940.0, 2920.0)])
def test_store_fit(region1, region2, peak, calfactor1, calfactor2):
    """
    Store fit
    Stored fit and active markers should stay fixed relative to spectrum
    """
    test_change_calibrated_fit(region1, region2, peak, calfactor1, calfactor2)
    spectra.StoreFit()
    markers_consistent(
        calfactor2, region1 / calfactor1 * calfactor2, region2 / calfactor1 * calfactor2
    )
    fit = list(spectra.Get("0").dict.items())[0][1]
    markers_consistent(
        calfactor2,
        region1 / calfactor1 * calfactor2,
        region2 / calfactor1 * calfactor2,
        fit=fit,
    )


@pytest.mark.parametrize("calfactor1", [2.1, 1.9])
@pytest.mark.parametrize("calfactor2", [0.4, 0.6])
@pytest.mark.parametrize("region1, region2, peak", [(2900.0, 2940.0, 2920.0)])
def test_reactivate_fit(region1, region2, peak, calfactor1, calfactor2):
    """
    Reactivate stored fit
    Active markers should appear at the same position as the stored fit
    """
    test_store_fit(region1, region2, peak, calfactor1, calfactor2)
    spectra.ActivateFit("0")
    markers_consistent(
        calfactor2, region1 / calfactor1 * calfactor2, region2 / calfactor1 * calfactor2
    )


@pytest.mark.parametrize("calfactor1", [2.1, 1.9])
@pytest.mark.parametrize("calfactor2", [0.4, 0.6])
@pytest.mark.parametrize("region1, region2, peak", [(2900.0, 2940.0, 2920.0)])
def test_execute_reactivated_fit(region1, region2, peak, calfactor1, calfactor2):
    """
    Execute reactivated fit
    New fit should be at the same position as stored fit
    """
    test_change_calibrated_fit(region1, region2, peak, calfactor1, calfactor2)
    spectra.StoreFit()
    out_original = list_fit()
    spectra.ActivateFit("0")
    test_reactivate_fit(region1, region2, peak, calfactor1, calfactor2)
    spectra.ExecuteFit()
    assert out_original == list_fit().replace("AV", " V")


@pytest.mark.parametrize("calfactor1", [2.1])
@pytest.mark.parametrize("calfactor2", [0.4])
@pytest.mark.parametrize("calfactor3", [1.9])
@pytest.mark.parametrize("region1, region2, peak", [(2900.0, 2940.0, 2920.0)])
def test_change_calibration_while_stored(
    region1, region2, peak, calfactor1, calfactor2, calfactor3
):
    """
    Change calibration after fit is stored
    Stored fit should move with spectrum to new position
    """
    test_execute_reactivated_fit(region1, region2, peak, calfactor1, calfactor2)
    spectra.ActivateFit(None)
    spectra.ClearFit()

    fit = list(spectra.Get("0").dict.items())[0][1]
    markers_initial = get_markers(fit)
    spectra.ApplyCalibration("0", [0, calfactor3])
    markers_final = get_markers(fit)

    assert isclose(markers_initial[1], markers_final[1] / calfactor3 * calfactor2)
    assert isclose(markers_initial[3], markers_final[3] / calfactor3 * calfactor2)
    assert isclose(markers_initial[5], markers_final[5] / calfactor3 * calfactor2)


@pytest.mark.parametrize("calfactor1", [2.1])
@pytest.mark.parametrize("calfactor2", [0.4])
@pytest.mark.parametrize("calfactor3", [1.9])
@pytest.mark.parametrize("region1, region2, peak", [(2900.0, 2940.0, 2920.0)])
def test_change_calibration_while_stored_activate(
    region1, region2, peak, calfactor1, calfactor2, calfactor3
):
    """
    Reactivate fit after calibration has changed
    Active markers should appear at the same position as fit
    """
    test_execute_reactivated_fit(region1, region2, peak, calfactor1, calfactor2)
    spectra.ActivateFit(None)
    spectra.ClearFit()
    spectra.ApplyCalibration("0", [0, calfactor3])

    fit = list(spectra.Get("0").dict.items())[0][1]
    markers_initial = get_markers(fit)
    spectra.ActivateFit("0")
    markers_final = get_markers()

    assert isclose(markers_initial[1], markers_final[1])
    assert isclose(markers_initial[3], markers_final[3])
    assert isclose(markers_initial[5], markers_final[5])


@pytest.mark.parametrize("calfactor1", [2.1])
@pytest.mark.parametrize("calfactor2", [0.4])
@pytest.mark.parametrize("calfactor3", [1.9])
@pytest.mark.parametrize("region1, region2, peak", [(2900.0, 2940.0, 2920.0)])
def test_execute_reactivated_calibrated_fit(
    region1, region2, peak, calfactor1, calfactor2, calfactor3
):
    """
    Execute reactivated fit
    New fit should be at the same position as the stored fit
    """
    test_execute_reactivated_fit(region1, region2, peak, calfactor1, calfactor2)
    spectra.ActivateFit(None)
    spectra.ClearFit()
    spectra.ApplyCalibration("0", [0, calfactor3])
    spectra.ActivateFit("0")

    spectra.ExecuteFit()

    fit = list(spectra.Get("0").dict.items())[0][1]
    markers_work = get_markers(fit)
    markers_stored = get_markers()

    assert isclose(markers_stored[1], markers_work[1], rel_tol=1e-7)
    assert isclose(markers_stored[3], markers_work[3], rel_tol=1e-7)
    assert isclose(markers_stored[5], markers_work[5], rel_tol=1e-7)


@pytest.mark.parametrize("calfactor1", [2.1])
@pytest.mark.parametrize("calfactor2", [0.4])
@pytest.mark.parametrize("region1, region2, peak", [(2900.0, 2940.0, 2920.0)])
def test_write_read_xml(temp_file, region1, region2, peak, calfactor1, calfactor2):
    """
    Write/read fit to/from xml
    Fit should appear at the same position as before
    """
    test_store_fit(region1, region2, peak, calfactor1, calfactor2)
    spectra.ActivateFit(None)

    out_original = list_fit()

    spectra.ClearFit()
    fitxml.WriteXML(spectra.Get("0").ID, temp_file)
    spectra.Get("0").Clear()
    fitxml.ReadXML(spectra.Get("0").ID, temp_file)

    assert out_original == list_fit()


@pytest.mark.parametrize("calfactor1", [1.9])
@pytest.mark.parametrize("calfactor2", [0.3])
@pytest.mark.parametrize("calfactor3", [0.7])
@pytest.mark.parametrize("region1, region2, peak", [(2900.0, 2940.0, 2920.0)])
def test_write_read_xml_calibrate(
    temp_file, region1, region2, peak, calfactor1, calfactor2, calfactor3
):
    """
    Write/Read fit and calibrate afterwards
    Fit should move with spectrum to new calibration
    """
    test_write_read_xml(temp_file, region1, region2, peak, calfactor1, calfactor2)

    fit = list(spectra.Get("0").dict.items())[0][1]
    markers_initial = get_markers(fit)
    integral_initial = fit.ExtractIntegralParams()[0][0]

    spectra.ApplyCalibration("0", [0, calfactor3])

    fit = list(spectra.Get("0").dict.items())[0][1]
    markers_final = get_markers(fit)
    integral_final = fit.ExtractIntegralParams()[0][0]

    assert isclose(markers_initial[1], markers_final[1] / calfactor3 * calfactor2)
    assert isclose(markers_initial[3], markers_final[3] / calfactor3 * calfactor2)
    assert isclose(markers_initial[5], markers_final[5] / calfactor3 * calfactor2)
    for key in ["pos", "width_cal"]:
        assert isclose(
            integral_initial[key].nominal_value,
            integral_final[key].nominal_value / calfactor3 * calfactor2,
        )


@pytest.mark.parametrize("calfactor1", [1.9])
@pytest.mark.parametrize("calfactor2", [0.3])
@pytest.mark.parametrize("calfactor3", [0.7])
@pytest.mark.parametrize("region1, region2, peak", [(2900.0, 2940.0, 2920.0)])
def test_reimport_calibrate_write_read(
    temp_file, region1, region2, peak, calfactor1, calfactor2, calfactor3
):
    """
    Write/Read fit and change calibration in between
    Fit should appear at the new calibrated position
    """
    test_store_fit(region1, region2, peak, calfactor1, calfactor2)
    spectra.ActivateFit(None)

    fit = list(spectra.Get("0").dict.items())[0][1]
    markers_initial = get_markers(fit)
    integral_initial = fit.ExtractIntegralParams()[0][0]

    fitxml.WriteXML(spectra.Get("0").ID, temp_file)
    spectra.Get("0").Clear()
    spectra.ApplyCalibration(0, [0, calfactor3])
    fitxml.ReadXML(spectra.Get("0").ID, temp_file)

    fit = list(spectra.Get("0").dict.items())[0][1]
    markers_final = get_markers(fit)
    integral_final = fit.ExtractIntegralParams()[0][0]

    assert isclose(markers_initial[1], markers_final[1] / calfactor3 * calfactor2)
    assert isclose(markers_initial[3], markers_final[3] / calfactor3 * calfactor2)
    assert isclose(markers_initial[5], markers_final[5] / calfactor3 * calfactor2)
    for key in ["pos", "width_cal"]:
        assert isclose(
            integral_initial[key].nominal_value,
            integral_final[key].nominal_value / calfactor3 * calfactor2,
        )



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from hdtv.drawable import DrawableManager
from hdtv.util import LockViewport


class Spectrum(DrawableManager):
    def __init__(self, histogram):
        # Create histogram attribute avoiding __setattr__ method
        self.__dict__["hist"] = histogram
        DrawableManager.__init__(self)

    # delegate everything to the underlying histogram
    def __setattr__(self, name, value):
        if self.hist is not None:
            self.hist.__setattr__(name, value)
        if hasattr(self, name):  # Update attribute of this class, if existent
            DrawableManager.__setattr__(self, name, value)

    def __getattr__(self, name):
        # The use of self.__dict__ replaces an infinite recusion by a KeyError
        #  if self.hist does not exist
        return getattr(self.__dict__["hist"], name)

    # color property
    def _set_color(self, color):
        for fit in self.dict.values():
            fit.color = color

    def _get_color(self):
        return self.hist.color

    color = property(_get_color, _set_color)

    # cal property
    def _set_cal(self, cal):
        for fit in self.dict.values():
            fit.cal = cal

    def _get_cal(self):
        return self.hist.cal

    cal = property(_get_cal, _set_cal)

    # overwrite some functions of DrawableManager to do some extra work
    def Insert(self, fit, ID=None):
        fit.spec = self
        return DrawableManager.Insert(self, fit, ID)

    def Pop(self, ID):
        fit = DrawableManager.Pop(self, ID)
        if fit is not None:
            fit.spec = None
        return fit

    def Draw(self, viewport):
        self.viewport = viewport
        with LockViewport(self.viewport):
            DrawableManager.Draw(self, viewport)
            if self.hist:
                self.hist.Draw(viewport)

    def Show(self):
        if self.viewport is None:
            return
        with LockViewport(self.viewport):
            DrawableManager.Show(self)
            if self.hist:
                self.hist.Show()

    def Hide(self):
        if self.viewport is None:
            return
        with LockViewport(self.viewport):
            DrawableManager.Hide(self)
            if self.hist:
                self.hist.Hide()

    def Refresh(self):
        if self.viewport is None:
            return
        with LockViewport(self.viewport):
            DrawableManager.Refresh(self)
            if self.hist:
                self.hist.Refresh()


class CutSpectrum(Spectrum):
    def __init__(self, hist, matrix, axis):
        Spectrum.__init__(self, hist)

        # Use self.__dict__ to avoid delegation to the underlying histogram
        self.__dict__["matrix"] = matrix
        self.__dict__["axis"] = axis

    def Show(self):
        """
        Show all visible cut markers for the axis of this spectrum
        """
        Spectrum.Show(self)
        if self.matrix:
            self.matrix.Show(axis=self.axis)

    def Hide(self):
        """
        Hide also the cut markers, but without changing the visibility state of the matrix
        """
        Spectrum.Hide(self)
        if self.matrix:
            self.matrix.Hide()

    def Refresh(self):
        """
        Repeat the cut
        """
        if self.matrix:
            for cut in self.matrix.dict.values():
                if self == cut.spec:
                    cut.Refresh()
                    break



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import os

import pytest

from tests.helpers.utils import redirect_stdout, hdtvcmd

import hdtv.cmdline
import hdtv.plugins.ls


@pytest.fixture(autouse=True)
def prepare(request):
    # original_wd = os.path.abspath(os.path.join(__file__, os.pardir))
    original_wd = os.getcwd()
    os.chdir(original_wd)
    yield
    os.chdir(original_wd)


def test_cmd_pwd():
    f, ferr = hdtvcmd("pwd")
    assert f == os.getcwd()


@pytest.mark.parametrize(
    "start, cd, target",
    [
        ("/", "/tmp", "/tmp"),
        ("/tmp", "..", "/"),
        ("/", "tmp", "/tmp"),
        ("/", "~", os.path.expanduser("~")),
        ("/tmp", "", os.path.expanduser("~")),
    ],
)
def test_cmd_cd(start, cd, target):
    os.chdir(start)
    hdtvcmd("cd " + cd)
    assert os.getcwd() == os.path.realpath(target)


def test_cmd_cd_minus():
    hdtvcmd("cd /")
    hdtvcmd("cd /tmp")
    hdtvcmd("cd -")
    assert os.getcwd() == "/"


@pytest.mark.parametrize(
    "cd, output", [("/tmp", "/tmp"), ("~", os.path.expanduser("~")), ("/", "/")]
)
def test_cmd_cd_output(cd, output):
    f, ferr = hdtvcmd("cd " + cd)
    assert f == output


def test_cmd_ls():
    os.chdir(os.path.abspath(os.sep))
    f, ferr = hdtvcmd("ls")
    assert "tmp" in f



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import copy
import sys
from html import escape

import ROOT
import hdtv.cmdline
import hdtv.options
import hdtv.util
import hdtv.ui
import hdtv.fit


class FitInterface(object):
    """
    User interface for fitting 1-d spectra
    """

    def __init__(self, spectra):
        hdtv.ui.debug("Loaded user interface for fitting of 1-d spectra")

        self.spectra = spectra
        self.window = self.spectra.window

        # tv commands
        self.tv = TvFitInterface(self)

        # Register configuration variables for fit interface
        # default region width for quickfit

        self.opt = dict()
        self.opt["quickfit.region"] = hdtv.options.Option(
            default=20.0, parse=lambda x: float(x)
        )
        hdtv.options.RegisterOption("fit.quickfit.region", self.opt["quickfit.region"])

        self.opt["display.decomp"] = hdtv.options.Option(
            default=False,
            parse=hdtv.options.parse_bool,
            changeCallback=lambda x: self.SetDecomposition(x),
        )
        hdtv.options.RegisterOption("fit.display.decomp", self.opt["display.decomp"])

        if self.window:
            self._register_hotkeys()

    def _register_hotkeys(self):
        self.window.AddHotkey(ROOT.kKey_b, lambda: self.spectra.SetMarker("bg"))
        self.window.AddHotkey(
            [ROOT.kKey_Minus, ROOT.kKey_b], lambda: self.spectra.RemoveMarker("bg")
        )
        self.window.AddHotkey(ROOT.kKey_r, lambda: self.spectra.SetMarker("region"))
        self.window.AddHotkey(
            [ROOT.kKey_Minus, ROOT.kKey_r], lambda: self.spectra.RemoveMarker("region")
        )
        self.window.AddHotkey(ROOT.kKey_p, lambda: self.spectra.SetMarker("peak"))
        self.window.AddHotkey(
            [ROOT.kKey_Minus, ROOT.kKey_p], lambda: self.spectra.RemoveMarker("peak")
        )
        self.window.AddHotkey(ROOT.kKey_B, lambda: self.spectra.ExecuteFit(peaks=False))
        self.window.AddHotkey(ROOT.kKey_F, lambda: self.spectra.ExecuteFit(peaks=True))
        self.window.AddHotkey(
            [ROOT.kKey_Minus, ROOT.kKey_B], lambda: self.spectra.ClearFit(bg_only=True)
        )
        self.window.AddHotkey(
            [ROOT.kKey_Minus, ROOT.kKey_F], lambda: self.spectra.ClearFit(bg_only=False)
        )
        self.window.AddHotkey(ROOT.kKey_Q, self.QuickFit)
        self.window.AddHotkey([ROOT.kKey_Plus, ROOT.kKey_F], self.spectra.StoreFit)
        self.window.AddHotkey([ROOT.kKey_Minus, ROOT.kKey_F], self.spectra.ClearFit)
        self.window.AddHotkey(ROOT.kKey_D, lambda: self.ShowDecomposition(True))
        self.window.AddHotkey(
            [ROOT.kKey_Minus, ROOT.kKey_D], lambda: self.ShowDecomposition(False)
        )
        self.window.AddHotkey(
            [ROOT.kKey_f, ROOT.kKey_s],
            lambda: self.window.EnterEditMode(
                prompt="Show Fit: ", handler=self._HotkeyShow
            ),
        )
        self.window.AddHotkey(
            [ROOT.kKey_f, ROOT.kKey_a],
            lambda: self.window.EnterEditMode(
                prompt="Activate Fit: ", handler=self._HotkeyActivate
            ),
        )
        self.window.AddHotkey(
            [ROOT.kKey_f, ROOT.kKey_p], lambda: self._HotkeyShow("PREV")
        )
        self.window.AddHotkey(
            [ROOT.kKey_f, ROOT.kKey_n], lambda: self._HotkeyShow("NEXT")
        )
        self.window.AddHotkey(ROOT.kKey_I, self.spectra.ExecuteIntegral)

    def _HotkeyShow(self, args):
        """
        Show wrapper for use with a Hotkey (internal use)
        """
        spec = self.spectra.GetActiveObject()
        if spec is None:
            self.window.viewport.SetStatusText("No active spectrum")
            return
        try:
            ids = hdtv.util.ID.ParseIds(args, spec)
        except ValueError:
            self.window.viewport.SetStatusText("Invalid fit identifier: %s" % args)
            return
        spec.ShowObjects(ids)

    def _HotkeyActivate(self, args):
        """
        ActivateObject wrapper for use with a Hotkey (internal use)
        """
        spec = self.spectra.GetActiveObject()
        if spec is None:
            self.window.viewport.SetStatusText("No active spectrum")
            return
        try:
            ids = hdtv.util.ID.ParseIds(args, spec)
        except ValueError:
            self.window.viewport.SetStatusText("Invalid fit identifier: %s" % args)
            return
        if len(ids) == 1:
            self.window.viewport.SetStatusText("Activating fit %s" % ids[0])
            self.spectra.ActivateFit(ids[0])
        elif not ids:
            self.window.viewport.SetStatusText("Deactivating fit")
            self.spectra.ActivateFit(None)
        else:
            self.window.viewport.SetStatusText("Can only activate one fit")

    def ExecuteRefit(self, specID, fitID, peaks=True):
        """
        Re-Execute Fit on store fits
        """
        try:
            spec = self.spectra.dict[specID]
        except KeyError:
            raise KeyError("invalid spectrum ID")
        try:
            fit = spec.dict[fitID]
        except KeyError:
            raise KeyError("invalid fit ID")
        if peaks:
            fit.FitPeakFunc(spec)
        else:
            if fit.fitter.backgroundModel.fParStatus["nparams"] == -1:
                raise RuntimeError("background degree of -1")
            fit.FitBgFunc(spec)
        hdtv.ui.msg(html=str(fit))
        fit.Draw(self.window.viewport)

    def ExecuteReintegrate(self, specID, fitID, print_result=True):
        """
        Re-Execute Fit on store fits
        """
        print("Reintegrating")
        try:
            spec = self.spectra.dict[specID]
        except KeyError:
            raise KeyError("invalid spectrum ID")
        try:
            fit = spec.dict[fitID]
        except KeyError:
            raise KeyError("invalid fit ID")
        if not fit.regionMarkers.IsFull():
            raise hdtv.cmdline.HDTVCommandAbort("Region not set.")
            return

        if fit.bgMarkers:
            if fit.fitter.backgroundModel.fParStatus["nparams"] == -1:
                raise hdtv.cmdline.HDTVCommandAbort(
                    "Background degree of -1 contradicts background fit."
                )
                return
            # pure background fit
            fit.FitBgFunc(spec)

        region = [fit.regionMarkers[0].p1.pos_uncal, fit.regionMarkers[0].p2.pos_uncal]
        bg = fit.fitter.bgFitter

        fit.integral = hdtv.integral.Integrate(spec, bg, region)
        if print_result:
            hdtv.ui.msg(html=fit.print_integral())
        fit.Draw(self.window.viewport)
        print("Successfully reintegrated")

    def QuickFit(self, pos=None):
        """
        Set region and peak markers automatically and do a quick fit as position "pos".

        If pos is not given, use cursor position
        """
        if pos is None:
            pos = self.window.viewport.GetCursorX()
        self.spectra.ClearFit()
        region_width = hdtv.options.Get("fit.quickfit.region")
        self.spectra.SetMarker("region", pos - region_width / 2.0)
        self.spectra.SetMarker("region", pos + region_width / 2.0)
        self.spectra.SetMarker("peak", pos)
        self.spectra.ExecuteFit()

    def ListFits(self, sid=None, ids=None, sortBy=None, reverseSort=False):
        """
        List results of stored fits as nice table
        """
        if sid is None:
            sid = self.spectra.activeID
        spec = self.spectra.dict[sid]
        # if there are not fits for this spectrum, there is not much to do
        if not spec.ids:
            hdtv.ui.msg("Spectrum " + str(sid) + " (" + spec.name + "): No fits")
            return
        # create result header
        result_header = "Fits in Spectrum " + str(sid) + " (" + spec.name + ")" + "\n"
        if ids is None:
            ids = spec.ids
        fits = [spec.dict[ID] for ID in ids]
        count_fits = len(fits)
        (objects, params) = self.ExtractFits(fits)

        # create result footer
        result_footer = (
            "\n" + str(len(objects)) + " peaks in " + str(count_fits) + " fits."
        )
        # create the table
        try:
            table = hdtv.util.Table(
                objects,
                params,
                sortBy=sortBy,
                reverseSort=reverseSort,
                extra_header=result_header,
                extra_footer=result_footer,
            )
            hdtv.ui.msg(html=str(table), end="")
        except KeyError as e:
            raise hdtv.cmdline.HDTVCommandError(
                "Spectrum " + str(sid) + ": No such attribute: " + str(e) + "\n"
                "Spectrum " + str(sid) + ": Valid attributes are: " + str(params)
            )

    def ListIntegrals(
        self, sid=None, ids=None, sortBy=None, reverseSort=False, integral_type="auto"
    ):
        """
        List integrals of fit regions as nice table
        """
        if sid is None:
            sid = self.spectra.activeID
        spec = self.spectra.dict[sid]
        # if there are not fits for this spectrum, there is not much to do
        if not spec.ids:
            hdtv.ui.msg("Spectrum " + str(sid) + " (" + spec.name + "): No fits")
            return
        # create result header
        result_header = (
            "Integrals of fit regions in Spectrum "
            + str(sid)
            + " ("
            + spec.name
            + ")"
            + "\n"
        )
        if ids is None:
            ids = spec.ids
        fits = [spec.dict[ID] for ID in ids]
        count_fits = len(fits)
        (objects, params) = self.ExtractIntegrals(fits, integral_type)

        # create the table
        try:
            table = hdtv.util.Table(
                objects,
                params,
                sortBy=sortBy,
                reverseSort=reverseSort,
                extra_header=result_header,
            )
            hdtv.ui.msg(html=str(table), end="")
        except KeyError as e:
            raise hdtv.cmdline.HDTVCommandError(
                "Spectrum " + str(sid) + ": No such attribute: " + str(e) + "\n"
                "Spectrum " + str(sid) + ": Valid attributes are: " + str(params)
            )

    def PrintWorkFit(self):
        """
        Print results of workFit as nice table
        """
        fit = self.spectra.workFit
        if fit.spec is not None:
            hdtv.ui.msg(html=str(fit))

    def PrintWorkFitIntegral(self):
        """
        Print integral of workFit range as nice table
        """
        # TODO
        fit = self.spectra.workFit
        if fit.spec is not None:
            hdtv.ui.msg(html=fit.print_integral())

    def ExtractFits(self, fits):
        """
        Helper function for use with ListFits and PrintWorkFit functions.

        Return values:
            fitlist    : a list of dicts for each peak in the fits
            params     : a ordered list of valid parameter names
        """
        fitlist = list()
        params = list()
        # loop through fits
        for fit in fits:
            # Get peaks
            (peaklist, fitparams) = fit.ExtractParams()
            if not peaklist:
                # If fit does not contain any peaks, fake an
                # entry to show that the fit ID is taken
                fakeID = hdtv.util.ID(fit.ID.major)
                # HACK: ID.minor is not allowed to contain strings,
                # but it will never be compared anyway and e.g. "0.-"
                # is nicer than just "0" in the table
                fakeID.minor = "-"
                fakeStat = "I"  # no peaks -> basically an integral only
                if fit.ID in fit.spec.visible:
                    fakeStat = fakeStat + "V"
                if fit.active:
                    fakeStat = "A" + fakeStat
                peaklist = [{"id": fakeID, "stat": fakeStat}]
                fitparams = ["id", "stat"]
            # update list of valid params
            for p in fitparams:
                # do not use set operations here to keep order of params
                if p not in params:
                    params.append(p)
            # add peaks to the list
            fitlist.extend(peaklist)
        return (fitlist, params)

    def ExtractIntegrals(self, fits, integral_type="auto"):
        """
        Helper function for functions ListIntegrals, PrintWorkFitIntegral.

        Return values:
            integrallist : a list of dicts for each integral in the fits
            params       : a ordered list of valid parameter names
        """
        integrallist = list()
        params = list()
        # loop through fits
        for fit in fits:
            (integrals, integralparams) = fit.ExtractIntegralParams(integral_type)

            # update list of valid params
            for p in integralparams:
                # do not use set operations here to keep order of params
                if p not in params:
                    params.append(p)
            # add peaks to the list
            integrallist.extend(integrals)
        return (integrallist, params)

    def ShowFitterStatus(self, ids=None):
        """
        Show status of the fit parameters of the work Fit.

        If default is true, the status of the default Fitter is shown in addition.
        If a list of ids is given, the status of the fitters belonging to that
        fits is also shown. Note, that the latter will silently fail for invalid
        IDs.
        """
        if ids is None:
            ids = list()
        ids.extend("a")
        statstr = str()
        for ID in ids:
            if ID == "a":
                fitter = self.spectra.workFit.fitter
                statstr += "active fitter: \n"
            else:
                spec = self.spectra.GetActiveObject()
                if spec is None:
                    continue
                fitter = spec.dict[ID].fitter
                statstr += "fitter status of fit id %d: \n" % ID
            if fitter.backgroundModel.name == "polynomial":
                statstr += "<b>Background model:</b> %s" % escape(
                    fitter.backgroundModel.name
                )
                statstr += ", deg=%i" % fitter.backgroundModel.fParStatus["nparams"]
                statstr += "\n"
            else:
                statstr += "<b>Background model:</b> %s\n" % escape(
                    fitter.backgroundModel.name
                )
            statstr += "<b>Peak model:</b> %s\n" % fitter.peakModel.name
            statstr += fitter.OptionsStr()
        hdtv.ui.msg(html=statstr)

    def SetFitterParameter(self, parname, status, ids=None):
        """
        Sets status of fitter parameter

        If not specified otherwise only the fitter of the workFit will be changed.
        If a list of fits is given, we try to set the parameter status also
        for these fits. Be aware that failures for the fits from the list will
        be silently ignored.
        """
        # active fit
        fit = self.spectra.workFit
        status = status.split(",")  # Create list from multiple stati
        if len(status) == 1:
            # Only single status was given so SetParameter need single string
            status = status[0]
        try:
            fit.fitter.SetParameter(parname, status)
            fit.Refresh()
        except ValueError as msg:
            raise hdtv.cmdline.HDTVCommandError(
                "while editing active Fit: \n\t%s" % msg
            )
        # fit list
        if not ids:  # works for None and empty list
            return
        spec = self.spectra.GetActiveObject()
        if spec is None:
            hdtv.ui.warning("No active spectrum")
        try:
            iter(ids)
        except TypeError:
            ids = [ids]
        for ID in ids:
            try:
                fit = spec.dict[ID]
                fit.fitter.SetParameter(parname, status)
                fit.Refresh()
            except KeyError:
                pass

    def ResetFitterParameters(self, ids=None):
        """
        Reset Status of Fitter

        The fitter of the workFit is resetted to the internal default values .
        If a list with ids is given, they are treated in the same way as the workFit.
        """
        # active Fit
        fit = self.spectra.workFit
        fit.fitter.ResetParamStatus()
        fit.Refresh()
        # fit list
        if not ids:  # works for None and empty list
            return
        spec = self.spectra.GetActiveObject()
        if spec is None:
            hdtv.ui.warning("No active spectrum")
        try:
            iter(ids)
        except TypeError:
            ids = [ids]
        for ID in ids:
            fit = spec.dict[ID]
            fit.fitter.ResetParamStatus()
            fit.Refresh()

    def SetPeakModel(self, peakmodel, ids=None):
        """
        Set the peak model (function used for fitting peaks)

        If a list of ids is given, the peak model of the stored fits with that
        ids will be set. Note, that the latter will fail silently for invalid
        fit ids.
        """
        # active fit
        fit = self.spectra.workFit
        fit.fitter.SetPeakModel(peakmodel)
        fit.Refresh()
        # fit list
        if not ids:  # works for None and empty list
            return
        spec = self.spectra.GetActiveObject()
        if spec is None:
            hdtv.ui.warning("No active spectrum")
        try:
            iter(ids)
        except TypeError:
            ids = [ids]
        for ID in ids:
            fit = spec.dict[ID]
            fit.fitter.SetPeakModel(peakmodel)
            fit.Refresh()

    def SetBackgroundModel(self, backgroundModel, ids=None):
        """
        Set the background model (function used for fitting the quasi-continuous background)

        If a list of ids is given, the peak model of the stored fits with that
        ids will be set. Note, that the latter will fail silently for invalid
        fit ids.
        """
        # active fit
        fit = self.spectra.workFit
        fit.fitter.SetBackgroundModel(backgroundModel)
        fit.Refresh()
        # fit list
        if not ids:  # works for None and empty list
            return
        spec = self.spectra.GetActiveObject()
        if spec is None:
            hdtv.ui.warning("No active spectrum")
        try:
            iter(ids)
        except TypeError:
            ids = [ids]
        for ID in ids:
            fit = spec.dict[ID]
            fit.fitter.SetBackgroundModel(backgroundModel)
            fit.Refresh()

    def SetDecomposition(self, default_enable):
        """
        Set default decomposition display status
        """
        # default_enable may be an hdtv.options.opt instance, so we
        # excplicitely convert to bool here
        default_enable = bool(default_enable)
        hdtv.fit.Fit.showDecomp = default_enable
        # show these decompositions for workFit
        self.ShowDecomposition(default_enable)
        # show these decompositions for all other fits
        for specID in self.spectra.ids:
            fitIDs = self.spectra.dict[specID].ids
            self.ShowDecomposition(default_enable, sid=specID, ids=fitIDs)

    def ShowDecomposition(self, enable, sid=None, ids=None):
        """
        Show decomposition of fits
        """

        fits = list()
        if sid is None:
            spec = self.spectra.GetActiveObject()
        else:
            spec = self.spectra.dict[sid]

        fits = list()
        if ids is None:
            if self.spectra.workFit is not None:
                fits.append(self.spectra.workFit)
        else:
            fits = [spec.dict[ID] for ID in ids]

        for fit in fits:
            fit.SetDecomp(enable)


class TvFitInterface(object):
    """
    TV style interface for fitting
    """

    def __init__(self, fitInterface):
        self.fitIf = fitInterface
        self.spectra = self.fitIf.spectra

        # Register configuration variables for fit list
        opt = hdtv.options.Option(default="ID")
        hdtv.options.RegisterOption("fit.list.sort_key", opt)

        prog = "fit execute"
        description = "(re)fit a fit"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="Spectra to work on",
        )
        parser.add_argument(
            "-b",
            "--background",
            action="store_true",
            default=False,
            help="fit only the background",
        )
        parser.add_argument(
            "-q",
            "--quick",
            action="store",
            default=None,
            type=float,
            help="set position for doing a quick fit",
        )
        parser.add_argument(
            "-S",
            "--store",
            action="store_true",
            default=False,
            help="store fit after fitting",
        )
        parser.add_argument(
            "fitids",
            nargs="*",
            default=None,
            help="id(s) of the fit(s) to (re)fit. Use 'none' to execute the WorkFit (default)",
        )
        hdtv.cmdline.AddCommand(prog, self.FitExecute, level=0, parser=parser)
        # the "fit execute" command is registered with level=0,
        # this allows "fit execute" to be abbreviated as "fit",
        # register all other commands starting with fit with default or higher
        # priority

        prog = "fit integral execute"
        description = "integrate over the fit region"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="Spectra to work on",
        )
        parser.add_argument(
            "-S",
            "--store",
            action="store_true",
            default=False,
            help="store integral (as fit without peaks) after integration",
        )
        parser.add_argument(
            "fitids",
            nargs="*",
            default=None,
            help="id(s) of the fit(s) to (re)integrate. Use 'none' to integrate the WorkFit (default)",
        )
        hdtv.cmdline.AddCommand(prog, self.FitIntegralExecute, parser=parser)

        prog = "fit marker"
        description = "set/delete a marker"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "type",
            # choices=['background', 'region', 'peak'], # no autocompletion
            help="type of marker to modify (background, region, peak)",
        )
        parser.add_argument(
            "action",
            # choices=['set', 'delete'], # no autocompletion
            help="set or delete marker",
        )
        parser.add_argument(
            "position", nargs="+", type=float, help="position of marker"
        )
        hdtv.cmdline.AddCommand(
            prog, self.FitMarkerChange, parser=parser, completer=self.MarkerCompleter
        )

        prog = "fit clear"
        description = "clear the active work fit"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-b",
            "--background_only",
            action="store_true",
            default=False,
            help="clear only background fit, refit peak fit with internal background",
        )
        hdtv.cmdline.AddCommand(prog, self.FitClear, parser=parser)

        prog = "fit store"
        description = "Store the current workFit"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "fitid", default=None, help="fitid to use for storage", nargs="?"
        )
        hdtv.cmdline.AddCommand(prog, self.FitStore, parser=parser)

        prog = "fit activate"
        description = "reactivates a fit from the fitlist"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "fitids",
            default=[],
            nargs="?",
            help="id of fit to reactivate. Use 'none' to activate the WorkFit, deactivating a fitlist fit, but keeping any fit markers (default)",
        )
        hdtv.cmdline.AddCommand(prog, self.FitActivate, parser=parser)

        prog = "fit delete"
        description = "delete fits"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="spectrum ids to work on",
        )
        parser.add_argument(
            "fitids", default=None, help="id(s) of fit(s) to delete", nargs="+"
        )
        hdtv.cmdline.AddCommand(prog, self.FitDelete, parser=parser)

        prog = "fit show"
        description = "display fits"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="select spectra to work on",
        )
        parser.add_argument(
            "-v",
            "--adjust-viewport",
            action="store_true",
            default=False,
            help="adjust viewport to include all fits",
        )
        parser.add_argument("fitids", default=None, help="id(s) of fit(s) to show")
        hdtv.cmdline.AddCommand(prog, self.FitShow, parser=parser)

        prog = "fit hide"
        description = "hide fits"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="select spectra to work on",
        )
        parser.add_argument("fitids", default=None, help="id(s) of fit(s) to hide")
        hdtv.cmdline.AddCommand(prog, self.FitHide, parser=parser)

        prog = "fit show decomposition"
        description = "display decomposition of fits"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="select spectra to work on",
        )
        parser.add_argument(
            "fitids",
            nargs="*",
            default=None,
            help="id(s) of fit(s) to show decomposition of. Use 'none' to show the decomposition of the WorkFit (default)",
        )
        hdtv.cmdline.AddCommand(prog, self.FitShowDecomp, parser=parser)

        prog = "fit hide decomposition"
        description = "display decomposition of fits"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="select spectra to work on",
        )
        parser.add_argument(
            "fitids",
            nargs="*",
            default=None,
            help="id(s) of fit(s) to hide decomposition of. Use 'none' to hide the decomposition of the WorkFit (default)",
        )
        hdtv.cmdline.AddCommand(prog, self.FitHideDecomp, parser=parser)

        prog = "fit focus"
        description = "adjust viewport to include all fits with id(s)"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s", "--spectrum", action="store", default="active", help="select spectra"
        )
        parser.add_argument("fitid", default=None, help="id(s) of fit(s) focus on")
        hdtv.cmdline.AddCommand(prog, self.FitFocus, parser=parser)

        prog = "fit list"
        description = "list fit results"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-v",
            "--visible",
            action="store_true",
            default=False,
            help="only list visible fit",
        )
        parser.add_argument(
            "-k", "--key-sort", action="store", default=None, help="sort by key"
        )
        parser.add_argument(
            "-r",
            "--reverse-sort",
            action="store_true",
            default=False,
            help="reverse the sort",
        )
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="select spectra to work on",
        )
        parser.add_argument(
            "-f",
            "--fit",
            action="store",
            default="all",
            help="specify which fits to list",
        )
        # FIXME: Why use a different syntax here? --fit vs fitids
        #       parser.add_argument(
        #           "fitids",
        #           default=None,
        #           help="id(s) of fit(s) to list",
        #           nargs='?')
        hdtv.cmdline.AddCommand(prog, self.FitList, parser=parser)

        prog = "fit integral list"
        description = "list integrals of fit ranges"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-v",
            "--visible",
            action="store_true",
            default=False,
            help="only list visible fit",
        )
        parser.add_argument(
            "-k",
            "--key-sort",
            action="store",
            default=hdtv.options.Get("fit.list.sort_key"),
            help="sort by key",
        )
        parser.add_argument(
            "-r",
            "--reverse-sort",
            action="store_true",
            default=False,
            help="reverse the sort",
        )
        parser.add_argument(
            "-i",
            "--integral-type",
            default="auto",
            help="Type of integral {auto,tot,bg,sub,all}. [Default=auto]",
        )
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="select spectra to work on",
        )
        parser.add_argument(
            "-f",
            "--fit",
            action="store",
            default="all",
            help="specify which fits to list",
        )
        # FIXME: Why use a different syntax here? --fit vs fitids
        #       parser.add_argument(
        #           "fitids",
        #           default=None,
        #           help="id(s) of fit(s) to list",
        #           nargs='?')
        hdtv.cmdline.AddCommand(prog, self.FitIntegralList, parser=parser)

        prog = "fit parameter"
        description = "show status of fit parameter, reset or set parameter"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-f",
            "--fit",
            action="store",
            default=None,
            help="change parameter of selected fit and refit. 'none' refers to the WorkFit (default)",
        )
        parser.add_argument(
            "action", help="{status,reset,background,tl,vol,pos,sh,sw,tr,width}"
        )
        parser.add_argument(
            "value_peak",
            metavar="value",
            help="fixed value to use for the parameter or instruction how to fit it \
                (e.g. 'free', 'equal', 'hold', 'none',...). Comma separated values can \
                be used to set values for each peak marker individually.",
            nargs="*",
        )
        hdtv.cmdline.AddCommand(
            prog, self.FitParam, completer=self.ParamCompleter, parser=parser
        )

        prog = "fit function peak activate"
        description = "selects which peak model to use"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-f",
            "--fit",
            action="store",
            default=None,
            help="change selected fits and refit",
        )
        parser.add_argument("peakmodel", help="name of peak model")
        hdtv.cmdline.AddCommand(
            prog, self.FitSetPeakModel, completer=self.PeakModelCompleter, parser=parser
        )

        prog = "fit function background activate"
        description = "selects which background model to use"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-f",
            "--fit",
            action="store",
            default=None,
            help="change selected fits and refit",
        )
        parser.add_argument("backgroundmodel", help="name of background model")
        hdtv.cmdline.AddCommand(
            prog,
            self.FitSetBackgroundModel,
            completer=self.BackgroundModelCompleter,
            parser=parser,
        )

    def FitMarkerChange(self, args):
        """
        Set or delete a marker from command line
        """
        # first argument is marker type name
        mtype = args.type
        # complete markertype if needed
        mtype = self.MarkerCompleter(mtype)
        if len(mtype) == 0:
            raise hdtv.cmdline.HDTVCommandError(
                "Markertype %s is not valid" % args.type
            )
        # second argument is action
        action = self.MarkerCompleter(
            args.action,
            args=[
                args.action,
            ],
        )
        if len(action) == 0:
            raise hdtv.cmdline.HDTVCommandError("Invalid action: %s" % args.action)
        # replace "background" with "bg" which is internally used
        mtype = mtype[0].strip()
        if mtype == "background":
            mtype = "bg"
        action = action[0].strip()

        # parse position
        for pos in args.position:
            if action == "set":
                self.spectra.SetMarker(mtype, pos)
            elif action == "delete":
                self.spectra.RemoveMarker(mtype, pos)

    def MarkerCompleter(self, text, args=[]):
        """
        Helper function for FitMarkerChange
        """
        if not args:
            mtypes = ["background", "region", "peak"]
            return hdtv.util.GetCompleteOptions(text, mtypes)
        elif len(args) == 1:
            actions = ["set", "delete"]
            return hdtv.util.GetCompleteOptions(text, actions)
        return []

    def FitExecute(self, args):
        """
        Execute a fit
        """
        specIDs = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        if len(specIDs) == 0:
            hdtv.ui.warning("No spectrum to work on")
            return
        if args.background:
            doPeaks = False
        else:
            doPeaks = True

        # Store active spec ID before activation of other spectra
        oldActiveID = self.spectra.activeID

        for specID in specIDs:
            self.spectra.ActivateObject(specID)
            fitIDs = hdtv.util.ID.ParseIds(args.fitids, self.spectra.dict[specID])
            if len(fitIDs) == 0:
                if args.quick is not None:
                    self.fitIf.QuickFit(args.quick)
                else:
                    self.spectra.ExecuteFit(peaks=doPeaks)

                # Needed when args.quick is set for multiple spectra, else fits will be lost
                if args.store is True:
                    self.spectra.StoreFit()  # Store current fit

            for fitID in fitIDs:
                try:
                    hdtv.ui.msg("Executing fit %s in spectrum %s" % (fitID, specID))
                    self.fitIf.ExecuteReintegrate(
                        specID=specID, fitID=fitID, print_result=False
                    )
                    self.fitIf.ExecuteRefit(specID=specID, fitID=fitID, peaks=doPeaks)
                except (KeyError, RuntimeError) as e:
                    hdtv.ui.warning(e)
                    continue

        if (
            oldActiveID is not None
        ):  # Reactivate spectrum that was active in the beginning
            self.spectra.ActivateObject(oldActiveID)
        return None

    def FitIntegralExecute(self, args):
        """
        Execute integral over fit region
        """
        specIDs = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        if not specIDs:
            hdtv.ui.warning("No spectrum to work on")
            return

        # Store active spec ID before activation of other spectra
        oldActiveID = self.spectra.activeID

        for specID in specIDs:
            self.spectra.ActivateObject(specID)
            fitIDs = hdtv.util.ID.ParseIds(args.fitids, self.spectra.dict[specID])
            if not fitIDs:
                self.spectra.ExecuteIntegral()

                if (
                    args.store is True
                ):  # Needed when args.quick is set for multiple spectra, else fits will be lost
                    self.spectra.StoreFit()  # Store current fit

            for fitID in fitIDs:
                try:
                    hdtv.ui.msg(
                        "Executing integral for region of fit %s in spectrum %s"
                        % (fitID, specID)
                    )
                    self.fitIf.ExecuteReintegrate(specID=specID, fitID=fitID)
                except (KeyError, RuntimeError) as e:
                    hdtv.ui.warning(e)
                    continue

        if (
            oldActiveID is not None
        ):  # Reactivate spectrum that was active in the beginning
            self.spectra.ActivateObject(oldActiveID)
        return None

    def FitClear(self, args):
        """
        Clear work fit
        """
        self.spectra.ClearFit(args.background_only)

    def FitStore(self, args):
        """
        Store work fit
        """
        try:
            self.spectra.StoreFit(args.fitid)
        except IndexError:
            return

    def FitActivate(self, args):
        """
        Activate a fit

        This marks a stored fit as active and copies its markers to the work Fit
        """
        sid = self.spectra.activeID
        if sid is None:
            hdtv.ui.warning("No active spectrum")
            return
        spec = self.spectra.dict[sid]
        ids = hdtv.util.ID.ParseIds(args.fitids, spec)
        if len(ids) == 1:
            hdtv.ui.msg("Activating fit %s" % ids[0])
            self.spectra.ActivateFit(ids[0], sid)
        elif not ids:
            hdtv.ui.msg("Deactivating fit")
            self.spectra.ActivateFit(None, sid)
        else:
            raise hdtv.cmdline.HDTVCommandError("Can only activate one fit")

    def FitDelete(self, args):
        """
        Delete fits
        """
        sids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        if not sids:
            hdtv.ui.warning("No spectra chosen or active")
            return
        else:
            for s in sids:
                spec = self.spectra.dict[s]
                fitids = hdtv.util.ID.ParseIds(args.fitids, spec, only_existent=False)
                already_removed = set()
                for fitid in fitids:
                    # only whole fits can be removed not single peaks
                    if fitid.minor is not None:
                        if fitid.major in already_removed:
                            continue
                        else:
                            msg = "It is not possible to remove single peaks, "
                            msg += (
                                "removing whole fit with id %s instead." % fitid.major
                            )
                            hdtv.ui.warning(msg)
                            fitid.minor = None
                            already_removed.add(fitid.major)
                    # do the work
                    spec.Pop(fitid)

    def FitHide(self, args):
        """
        Hide Fits
        """
        # FitHide is the same as FitShow, except that the spectrum selection is
        # inverted
        self.FitShow(args, inverse=True)

    def FitShow(self, args, inverse=False):
        """
        Show Fits

        inverse = True inverses the fit selection i.e. FitShow becomes FitHide
        """
        sids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        for sid in sids:
            spec = self.spectra.dict[sid]
            fitids = hdtv.util.ID.ParseIds(args.fitids, spec)
            if inverse:
                spec.HideObjects(fitids)
            else:
                spec.ShowObjects(fitids)
                if args.adjust_viewport:
                    fits = [spec.dict[fitid] for fitid in fitids]
                    self.spectra.window.FocusObjects(fits)

    def FitHideDecomp(self, args):
        """
        Hide decomposition of fits
        """
        self.FitShowDecomp(args, show=False)

    def FitShowDecomp(self, args, show=True):
        """
        Show decomposition of fits

        show = False hides decomposition
        """
        sids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        for sid in sids:
            spec = self.spectra.dict[sid]
            fitIDs = hdtv.util.ID.ParseIds(args.fitids, spec)

            if not fitIDs:
                fitIDs = None
            self.fitIf.ShowDecomposition(show, sid=sid, ids=fitIDs)

    def FitFocus(self, args):
        """
        Focus a fit.

        If no fit is given focus the active fit.
        """
        sids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)

        fits = list()
        if not args.fitid:
            fits.append(self.spectra.workFit)
            spec = self.spectra.GetActiveObject()
            activeFit = spec.GetActiveObject()
            if activeFit is not None:
                fits.append(activeFit)
        else:
            for sid in sids:
                spec = self.spectra.dict[sid]
                ids = hdtv.util.ID.ParseIds(args.fitid, spec)
                fits.extend([spec.dict[ID] for ID in ids])
                spec.ShowObjects(ids, clear=False)
                if not fits:
                    hdtv.ui.warning("Nothing to focus in spectrum %s" % sid)
                    return
        self.spectra.window.FocusObjects(fits)

    def FitList(self, args):
        """
        Show a nice table with the results of fits

        By default the result of the work fit is shown.
        """
        self.fitIf.PrintWorkFit()
        sids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        if not sids:
            hdtv.ui.warning("No spectra chosen or active")
            return
        # parse sort_key
        if args.key_sort is None:
            args.key_sort = hdtv.options.Get("fit.list.sort_key")
        key_sort = args.key_sort.lower()
        for sid in sids:
            spec = self.spectra.dict[sid]
            ids = hdtv.util.ID.ParseIds(args.fit, spec)
            if args.visible:
                ids = [ID for ID in spec.visible]
            if not ids:
                continue
            self.fitIf.ListFits(
                sid, ids, sortBy=key_sort, reverseSort=args.reverse_sort
            )

    def FitIntegralList(self, args):
        """
        Show a nice table with the results of fits

        By default the result of the work fit is shown.
        """
        self.fitIf.PrintWorkFitIntegral()
        sids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        if not sids:
            hdtv.ui.warning("No spectra chosen or active")
            return
        # parse sort_key
        key_sort = args.key_sort.lower()
        for sid in sids:
            spec = self.spectra.dict[sid]
            ids = hdtv.util.ID.ParseIds(args.fit, spec)
            if args.visible:
                ids = [ID for ID in spec.visible]
            if not ids:
                continue
            self.fitIf.ListIntegrals(
                sid,
                ids,
                sortBy=key_sort,
                reverseSort=args.reverse_sort,
                integral_type=args.integral_type,
            )

    def FitSetPeakModel(self, args):
        """
        Defines peak model to use for fitting
        """
        name = args.peakmodel.lower()
        # complete the model name if needed
        models = self.PeakModelCompleter(name)
        # check for unambiguity
        if len(models) > 1:
            raise hdtv.cmdline.HDTVCommandError(
                "Peak model name '%s' is ambiguous" % name
            )
        if not models:
            raise hdtv.cmdline.HDTVCommandError("Invalid peak model '%s'" % name)
        else:
            name = models[0].strip()
            ids = list()
            if args.fit:
                spec = self.spectra.GetActiveObject()
                if spec is None:
                    hdtv.ui.warning("No active spectrum, no action taken.")
                    return
                ids = hdtv.util.ID.ParseIds(args.fit, spec)
            self.fitIf.SetPeakModel(name, ids)

    def FitSetBackgroundModel(self, args):
        """
        Defines background model to use for fitting
        """
        name = args.backgroundmodel.lower()
        # complete the model name if needed
        models = self.BackgroundModelCompleter(name)
        # check for unambiguity
        if len(models) > 1:
            raise hdtv.cmdline.HDTVCommandError(
                "Background model name '%s' is ambiguous" % name
            )
        if not models:
            raise hdtv.cmdline.HDTVCommandError("Invalid background model '%s'" % name)
        else:
            name = models[0].strip()
            ids = list()
            if args.fit:
                spec = self.spectra.GetActiveObject()
                if spec is None:
                    hdtv.ui.warning("No active spectrum, no action taken.")
                    return
                ids = hdtv.util.ID.ParseIds(args.fit, spec)
            self.fitIf.SetBackgroundModel(name, ids)

    def PeakModelCompleter(self, text, args=None):
        """
        Helper function for FitSetPeakModel
        """
        return hdtv.util.GetCompleteOptions(
            text, iter(hdtv.peakmodels.PeakModels.keys())
        )

    def BackgroundModelCompleter(self, text, args=None):
        """
        Helper function for FitSetBackgroundModel
        """
        return hdtv.util.GetCompleteOptions(
            text, iter(hdtv.backgroundmodels.BackgroundModels.keys())
        )

    def FitParam(self, args):
        """
        Manipulate the status of fitter parameter
        """
        # first argument is parameter name
        param = args.action
        # complete the parameter name if needed
        parameter = self.ParamCompleter(param)
        # check for unambiguity
        if len(parameter) > 1:
            raise hdtv.cmdline.HDTVCommandError(
                "Parameter name %s is ambiguous" % param
            )
        if not parameter:
            raise hdtv.cmdline.HDTVCommandError(
                "Parameter name %s is not valid" % param
            )
        param = parameter[0].strip()
        ids = list()
        if args.fit:
            spec = self.spectra.GetActiveObject()
            if spec is None:
                hdtv.ui.warning("No active spectrum, no action taken.")
                return
            ids = hdtv.util.ID.ParseIds(args.fit, spec)
        if param == "status":
            self.fitIf.ShowFitterStatus(ids)
        elif param == "reset":
            self.fitIf.ResetFitterParameters(ids)
        else:
            try:
                self.fitIf.SetFitterParameter(param, " ".join(args.value_peak), ids)
            except ValueError as msg:
                raise hdtv.cmdline.HDTVCommandError(msg)

    def ParamCompleter(self, text, args=None):
        """
        Creates a completer for all possible parameter names
        or valid states for a parameter (args[0]: parameter name).
        """
        if not args:  # args is None or [] -> complete parameter names
            params = ["status", "reset"]
            # create a list of all possible parameter names
            params.extend(self.spectra.workFit.fitter.params)
            return hdtv.util.GetCompleteOptions(text, params)
        else:  # args[0] = parameter name -> complete its possible values
            states = list()
            param = args[0]
            if param in ["status", "reset", "background"]:
                return []  # No further args to autocomplete
            else:
                activePM = self.spectra.workFit.fitter.peakModel
                try:
                    if param in activePM.fValidParStatus:
                        valid_status = activePM.fValidParStatus[param]
                    else:
                        valid_status = activePM.fValidOptStatus[param]
                    states = [
                        str(s).lower()
                        for s in valid_status
                        if isinstance(s, (str, bool))
                    ]
                except KeyError:
                    # param is not a parameter of the peak model of active
                    # fitter
                    msg = "Invalid parameter %s for active peak model %s" % (
                        param,
                        activePM.name,
                    )
                    raise hdtv.cmdline.HDTVCommandAbort(msg)
                return hdtv.util.GetCompleteOptions(text, states)

    def ResetFit(self, args):
        """
        Reset fitter of a fit to unfitted default.
        """
        specIDs = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        if not specIDs:
            raise hdtv.cmdline.HDTVCommandError("No spectrum to work on")
        for specID in specIDs:
            fitIDs = hdtv.util.ID.ParseIds(args.fitids, self.spectra.dict[specID])
            if not fitIDs:
                hdtv.ui.warning("No fit for spectrum %d to work on" % specID)
                continue
            for fitID in fitIDs:
                self.fitIf.FitReset(
                    specID=specID, fitID=fitID, resetFitter=not args.keep_fitter
                )


# plugin initialisation

import __main__

fit_interface = FitInterface(__main__.spectra)
hdtv.cmdline.RegisterInteractive("f", fit_interface)



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import pytest

from tests.helpers.utils import redirect_stdout, hdtvcmd

from hdtv.util import monkey_patch_ui

monkey_patch_ui()

import hdtv.cmdline
import hdtv.options
import hdtv.window
import hdtv.session

import __main__

try:
    __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass


import hdtv.plugins.specInterface
import hdtv.plugins.calInterface
import hdtv.plugins.fitInterface
import hdtv.plugins.peakfinder
import hdtv.plugins.fitmap
import hdtv.plugins.config
import hdtv.plugins.dblookup
import hdtv.plugins.fittex
import hdtv.plugins.matInterface
import hdtv.plugins.rootInterface
import hdtv.plugins.run
import hdtv.plugins.fitlist
import hdtv.plugins.printing

cmdlist = [
    "calibration efficiency fit",
    "calibration efficiency list",
    "calibration efficiency plot",
    "calibration efficiency read covariance",
    "calibration efficiency read parameter",
    "calibration efficiency set",
    "calibration efficiency write covariance",
    "calibration efficiency write parameter",
    "calibration position assign",
    "calibration position copy",
    "calibration position enter",
    "calibration position list",
    "calibration position list clear",
    "calibration position list read",
    "calibration position list write",
    "calibration position nuclide",
    "calibration position read",
    "calibration position recalibrate",
    "calibration position set",
    "calibration position unset",
    "config reset",
    "config set",
    "config show",
    "cut activate",
    "cut clear",
    "cut delete",
    "cut execute",
    "cut hide",
    "cut marker",
    "cut show",
    "cut store",
    "db info",
    "db list",
    "db lookup",
    "db set",
    "fit activate",
    "fit clear",
    "fit delete",
    "fit execute",
    "fit focus",
    "fit function peak activate",
    "fit getlists",
    "fit hide",
    "fit hide decomposition",
    "fit integral list",
    "fit integral execute",
    "fit list",
    "fit marker",
    "fit parameter",
    "fit peakfind",
    "fit position assign",
    "fit position erase",
    "fit position map",
    "fit read",
    "fit savelists",
    "fit show",
    "fit show decomposition",
    "fit store",
    "fit tex",
    "fit write",
    # "matrix delete",
    "matrix get",
    "matrix list",
    # "matrix project",
    "matrix view",
    "nuclide",
    "print",
    "root cut delete",
    "root cut view",
    "root get",
    "root matrix get",
    "root matrix view",
    "spectrum activate",
    "spectrum add",
    "spectrum calbin",
    "spectrum copy",
    "spectrum delete",
    "spectrum get",
    "spectrum hide",
    "spectrum info",
    "spectrum list",
    "spectrum multiply",
    "spectrum name",
    "spectrum normalize",
    "spectrum rebin",
    "spectrum show",
    "spectrum substract",
    "spectrum update",
    "spectrum write",
    "window view center",
    "window view region",
]


@pytest.mark.parametrize("command", cmdlist)
def test_cmd_help(command):
    f, ferr = hdtvcmd("{} --help".format(command))
    assert ferr == ""
    assert "usage" in f
    assert "--help" in f


@pytest.mark.parametrize("command", cmdlist)
def test_cmd_unrecognized_arg(command):
    f, ferr = hdtvcmd("{} --invalidarg".format(command))
    assert "ERROR" in ferr
    assert "usage" in f



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from .wiedenhoever import WiedenhoeverEff
from .wunder import WunderEff
from .polyeff import PolyEff
from .exp import ExpEff
from .pow import PowEff

# from . orthogonal import OrthogonalEff



# -*- coding: utf-8 -*-
import csv
import os
from uncertainties import ufloat_fromstr
import hdtv.cmdline
import hdtv.ui


class _Element(object):
    """
    Store info about elements
    """

    __slots__ = ("ID", "z", "symbol", "name", "m")

    def __init__(self, Z, Symbol, Name=None, M=None):

        self.z = Z
        self.symbol = Symbol  # Symbol of element (e.g. H)
        self.name = Name  # Full name of element (e.g. Hydrogen)
        # Atomic weight, This is normalizes over nuclides of element and
        # abundancies
        self.m = M
        self.ID = self.symbol

    def _get_Z(self):
        return self.z

    Z = property(_get_Z)

    def _get_M(self):
        return self.m

    M = property(_get_M)

    def __str__(self):
        text = str()
        text += "Name:          " + str(self.name) + "\n"
        text += "Z:             " + str(self.z) + " \n"
        text += "Symbol:        " + self.symbol + " \n"
        text += "atomic Mass:   " + str(self.m) + " u \n"
        return text


class _Elements(list):
    """
    Read and hold complete elements list
    """

    def __init__(self, csvfile=os.path.join(hdtv.datadir, "elements.dat")):

        super(_Elements, self).__init__()

        tmp = list()

        try:
            try:
                datfile = open(csvfile, "r", encoding="utf-8")
            except TypeError:
                datfile = open(csvfile, "rb")
            reader = csv.reader(datfile)
            next(reader)  # Skip header

            for line in reader:
                Z = int(line[0])
                Symbol = line[1].strip()
                Name = line[2].strip()
                try:
                    Mass = ufloat_fromstr(line[3].strip())
                except ValueError:
                    Mass = None
                element = _Element(Z, Symbol, Name, Mass)
                tmp.append(element)
        except csv.Error as err:
            hdtv.ui.error("file %s, line %d: %s" % (csvfile, reader.line_num, err))
        finally:
            datfile.close()

        # Now store elements finally
        maxZ = max(tmp, key=lambda x: x.z)  # Get highest Z
        for i in range(maxZ.z):
            self.append(None)

        for e in tmp:
            self[e.z] = e

    def __call__(self, Z=None, symbol=None, name=None):

        if symbol:
            for z in self:
                try:
                    if z.symbol.lower() == symbol.lower():
                        return z
                except AttributeError:
                    pass
            raise ValueError  # If we reach this point there is no element with this symbol
        if name:
            for z in self:
                try:
                    if z.name.lower() == name.lower():
                        return z
                except AttributeError:
                    pass
            raise ValueError  # If we reach this point there is no element with this name

        if Z:
            return self[Z]

        return self

    def __setitem__(self, index, value):
        if index == 0:
            return None

        if index > 0:
            index = index - 1

        return super(_Elements, self).__setitem__(index, value)

    def __getitem__(self, index):
        if index == 0:
            return None

        if index > 0:
            index = index - 1

        return super(_Elements, self).__getitem__(index)


class _Nuclide(_Element):

    __slots__ = ("ID", "element", "a", "m", "sigma", "abundance")

    def __init__(self, element, A, abundance=None, sigma=None, M=None):

        self.element = element
        super(_Nuclide, self).__init__(
            element.z, element.symbol, element.name, M=None
        )  # M is different for nuclides and elements
        self.a = A
        self.m = M
        self.sigma = sigma
        self.abundance = abundance
        self.ID = str(self.a) + "-" + str(self.symbol)

    def _get_symbol(self):
        return self.element.symbol

    def _set_symbol(self, symbol):
        self.element.symbol = symbol

    symbol = property(_get_symbol, _set_symbol)  # def _Z(self):

    def __str__(self):
        text = str()
        text += "ID:            " + self.ID + "\n"
        text += "Z:             " + str(self.z) + " \n"
        text += "A:             " + str(self.a) + "\n"
        text += "atomic Mass:   " + str(self.m) + " u\n"
        try:
            abundance = str(self.abundance * 100.0) + " %"
        except TypeError:  # Unstable nuclide -> no abundance
            abundance = "---"
        text += "Abundance:     " + abundance + " \n"
        if self.sigma.nominal_value is None:
            sigma = "???"
        else:
            sigma = str(self.sigma) + " b"
        text += "Sigma_0:       " + sigma + "\n"
        return text


class _Nuclides(object):
    def __init__(self, csvfile=os.path.join(hdtv.datadir, "nuclides.dat")):

        self._storage = dict()
        try:
            try:
                datfile = open(csvfile, "r", encoding="utf-8")
            except TypeError:
                datfile = open(csvfile, "rb")
            reader = csv.reader(datfile)
            next(reader)  # Skip header

            for line in reader:
                Z = int(line[0].strip())
                element = Elements(Z)
                A = int(line[1].strip())
                if line[2].strip():
                    abd = ufloat_fromstr(line[2].strip()) / 100.0
                else:
                    abd = None
                if line[3].strip():
                    M = ufloat_fromstr(line[3].strip())
                else:
                    M = None
                if line[4].strip():
                    sigma = ufloat_fromstr(line[4].strip())
                else:
                    sigma = None

                if Z not in self._storage:
                    self._storage[Z] = dict()
                self._storage[Z][A] = _Nuclide(
                    element, A, abundance=abd, sigma=sigma, M=M
                )
        except csv.Error as e:
            hdtv.ui.error("file %s, line %d: %s" % (csvfile, reader.line_num, e))
        finally:
            datfile.close()

    def __call__(self, Z=None, A=None, symbol=None, name=None):
        """
        Return a list of nuclides with given properties (Z, A, symbol, name)

        e.g.: Nuclides(A=197, symbol="Au") returns [Au-197]
              Nuclides(symbol="Au") or Nuclides(name="gold") or Nuclides(Z=79) return list of all gold nuclides
        """

        ret = list()

        if (
            Z is not None
        ):  # Z is given, so it is faster if we start with nuclides of this Z
            if A is not None:  # Nuclide uniquely defined
                ret.append(self._storage[Z][A])
            else:  # All nuclides with given Z
                for n in self._storage[Z].values():
                    ret.append(n)

        if len(ret) == 0:  # Z was not given, so we have to cycle through all nuclides
            for n in self._storage.values():
                for e in n.values():
                    ret.append(e)

        tmp = list()

        # Select by A
        if A is not None:
            for n in ret:
                if n.a == A:
                    tmp.append(n)

            ret = tmp
            tmp = list()

        # Select by symbol
        if symbol is not None:
            for n in ret:
                if n.symbol.lower() == symbol.lower():
                    tmp.append(n)

            ret = tmp
            tmp = list()

        # Select by name
        if name is not None:
            for n in ret:
                if n.name.lower() == name.lower():
                    tmp.append(n)

            ret = tmp

        return ret


class Gamma(object):
    """Class for storing information about gammas"""

    __slots__ = ("ID", "nuclide", "energy", "sigma", "intensity")

    def __init__(self, nuclide, energy, sigma, intensity):
        self.ID = str(nuclide.ID) + "@" + str(energy.nominal_value.__int__())
        self.nuclide = nuclide
        self.energy = energy
        self.sigma = sigma
        self.intensity = intensity

    def __str__(self):
        text = str()
        text += "ID:        " + str(self.ID) + "\n"
        text += "Energy:    " + str(self.energy) + " keV\n"
        if self.sigma is not None:
            text += "Sigma:     " + str(self.sigma) + " b\n"
        if self.intensity is not None:
            text += "Intensity: " + str(self.intensity * 100.0) + " %\n"
        return text

    def _Z(self):
        return self.nuclide.z

    z = property(_Z)

    def _A(self):
        return self.nuclide.a

    a = property(_A)

    def _symbol(self):
        return self.nuclide.symbol

    symbol = property(_symbol)

    def __eq__(self, other):
        if self.energy == other.energy and other.ID is not None:
            if self.ID == other.ID:
                return self.energy == other.energy
            else:
                return self.ID == other.ID
        else:
            return self.energy == other.energy

    def __ne__(self, other):
        if self.energy == other.energy and other.ID is not None:
            if self.ID == other.ID:
                return self.energy != other.energy
            else:
                return self.ID != other.ID
        else:
            return self.energy != other.energy

    def __gt__(self, other):
        if self.energy == other.energy and other.ID is not None:
            if self.ID == other.ID:
                return self.energy < other.energy
            else:
                return self.ID < other.ID
        else:
            return self.energy < other.energy

    def __lt__(self, other):
        if self.energy == other.energy and other.ID is not None:
            if self.ID == other.ID:
                return self.energy > other.energy
            else:
                return self.ID > other.ID
        else:
            return self.energy > other.energy

    def __ge__(self, other):
        if self.energy == other.energy and other.ID is not None:
            if self.ID == other.ID:
                return self.energy >= other.energy
            else:
                return self.ID >= other.ID
        else:
            return self.energy >= other.energy

    def __le__(self, other):
        if self.energy == other.energy and other.ID is not None:
            if self.ID == other.ID:
                return self.energy <= other.energy
            else:
                return self.ID <= other.ID
        else:
            return self.energy <= other.energy


# TODO: this should be an abstract baseclass!
class GammaLib(list):
    """
    Class for storing a gamma library

    The real libs should be derived from this
    """

    __slots__ = ("nuclide", "energy", "sigma", "intensity", "E_fuzziness")

    def __init__(self, fuzziness=1.0):
        list.__init__(self)
        self.fuzziness = fuzziness  # Fuzzyness for energy identification
        self.opened = False

    def find(self, fuzziness=None, sort_key=None, sort_reverse=False, **args):
        """
        Find in gamma lib

        Does a fuzzy compare for floats. All strings are compared lowercase.

        Valid key args are:

         * sort_key: key to sort
         * sort_reverse: sort_reverse
         * "key: value" : key value pairs to find
        """
        if not self.opened:
            self.open()

        if fuzziness is None:
            fuzziness = self.fuzziness

        # convert keys to lowercase
        fields_lower = dict()
        args_lower = dict()
        for (key, conv) in list(self.fParamConv.items()):
            fields_lower[key.lower()] = conv

        for (key, value) in list(args.items()):
            conv = fields_lower[key.lower()]
            if value is None:
                continue
            args_lower[key.lower()] = conv(value)

        # Prepare find args
        fargs = dict()

        for (key, conv) in list(fields_lower.items()):
            try:
                # Convert as described in fParamConv dict
                fargs[key] = conv(args_lower[key])
            except KeyError:
                pass

        # Do the search
        results = self[:]

        if fuzziness is None:
            fuzziness = self.fuzziness

        if not fargs:
            return []

        for (key, value) in list(fargs.items()):
            if value is None:
                continue
            if isinstance(value, int):
                results = [x for x in results if getattr(x, key) == value]
            elif isinstance(value, str):  # Do lowercase comparison for strings
                results = [
                    x for x in results if getattr(x, key).lower() == value.lower()
                ]
            else:  # Do fuzzy compare
                results = [
                    x for x in results if abs(getattr(x, key) - value) <= fuzziness
                ]

        # Sort
        try:
            if sort_key is not None:
                results.sort(key=lambda x: getattr(x, sort_key), reverse=sort_reverse)
        except AttributeError:
            hdtv.ui.warning("Could not sort by '" + str(sort_key) + "': No such key")
            raise AttributeError

        return results


Elements = _Elements()
Nuclides = _Nuclides()



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Peak finding and fitting plugin for HDTV
"""

import copy

import hdtv.cmdline
import hdtv.options
import hdtv.ui
import hdtv.plugins

import ROOT


class PeakFinder(object):
    """
    Automatic peak finder - using ROOTS peak search function
    """

    def __init__(self, spectra):
        self.spectra = spectra
        hdtv.ui.debug("Loaded PeakFinder plugin")

    def __call__(
        self, sid, sigma, threshold, start=None, end=None, autofit=False, reject=False
    ):
        self.spec = self.spectra.dict[sid]
        self.sigma_E = sigma
        peaks = self.PeakSearch(sigma, threshold, start, end)
        num = self.StoreFits(peaks, autofit, reject)
        hdtv.ui.msg("Found " + str(num) + " peaks")
        # remove reference to spec otherwise we get trouble with garbage
        # collection
        self.spec = None

    def PeakSearch(self, sigma, threshold, start=None, end=None):
        """
        Search for peaks
        """
        sigma_E = sigma

        tSpec = ROOT.TSpectrum()

        # Copy underlying ROOT hist here so we can safely modify ranges, etc.
        # below
        hist = copy.copy(self.spec.hist).hist

        # Init start and end region
        if start is None:
            start_E = 0.0
        else:
            start_E = start
        start_Ch = self.spec.cal.E2Ch(start_E)
        if end is None:
            end_Ch = hist.GetNbinsX()
            end_E = self.spec.cal.Ch2E(end_Ch)
        else:
            end_E = end
            end_Ch = self.spec.cal.E2Ch(end_E)

        # good approximation of sigma_Ch
        sigma_Ch = self.spec.cal.E2Ch(sigma) - self.spec.cal.E2Ch(0.0)
        assert sigma_Ch > 0, "Sigma must be > 0"

        text = "Search Peaks in region "
        text += str(start_E) + "--" + str(end_E)
        text += " (sigma=" + str(sigma_E)
        text += " threshold=" + str(threshold * 100) + "%)"
        hdtv.ui.msg(text)

        # Invoke ROOT's peak finder
        hist.SetAxisRange(start_Ch, end_Ch)
        num_peaks = tSpec.Search(hist, sigma_Ch, "goff", threshold)
        foundpeaks = tSpec.GetPositionX()

        # Sort by position
        tmp = list()
        for i in range(0, num_peaks):
            tmp.append(foundpeaks[i])  # convert from array to list
        tmp.sort()
        foundpeaks = tmp

        return foundpeaks

    def StoreFits(self, foundpeaks, autofit=False, reject=False):
        """
        Create fit objects from peak positions and add them to the fitlist
        If autofit is set to True fitting is done,
        if reject is set to True all badFits will be remove.
        """
        peak_count = 0
        while len(foundpeaks) > 0:
            p = foundpeaks.pop(0)
            fitter = copy.copy(self.spectra.workFit.fitter)
            fit = hdtv.fit.Fit(fitter, cal=self.spec.cal)
            pos_E = self.spec.cal.Ch2E(p)
            fit.ChangeMarker("peak", pos_E, action="set")
            if autofit:
                region_width = self.sigma_E * 5.0  # TODO: something sensible here
                # left region marker
                fit.ChangeMarker("region", pos_E - region_width / 2.0, action="set")
                limit = pos_E + region_width
                # collect multipletts
                while (
                    len(foundpeaks) > 0 and self.spec.cal.Ch2E(foundpeaks[0]) <= limit
                ):
                    next = foundpeaks.pop(0)
                    pos_E = self.spec.cal.Ch2E(next)
                    limit = pos_E + region_width
                    fit.ChangeMarker("peak", pos_E, "set")
                # right region marker
                fit.ChangeMarker("region", pos_E + region_width / 2.0, "set")
                fit.FitPeakFunc(self.spec)  # , silent = True
                # check fits
                result = self.BadFit(fit)
                if reject:
                    if result:
                        text = "Rejecting invalid fit:" + result
                        hdtv.ui.msg(text)
                        continue
                else:
                    if result:
                        text = "Adding invalid fit:" + result
                        hdtv.ui.warning(text)
            # Integrate. TODO: Might use this for additional checks
            if fit.regionMarkers.IsFull():
                region = [
                    fit.regionMarkers[0].p1.pos_uncal,
                    fit.regionMarkers[0].p2.pos_uncal,
                ]
                fit.integral = hdtv.integral.Integrate(
                    self.spec, fit.fitter.bgFitter, region
                )
            # add fits to spectrum
            ID = self.spec.Insert(fit)
            # FIXME: no fit title
            # fit.title = fit.title + "(*)"
            # bookkeeping
            if len(fit.peaks) > 0:
                peak_count = peak_count + len(fit.peaks)
            else:
                peak_count = peak_count + 1

        return peak_count

    def BadFit(self, fit):
        """
        Check if the fit is sensible
        """
        bad = False
        text = str()
        for peak in fit.peaks:
            reason = None
            if peak.vol.nominal_value <= 0.0:
                bad = True
                reason = "vol = %s" % peak.vol
            elif (
                not fit.regionMarkers[0].p1.pos_cal
                < peak.pos_cal.nominal_value
                < fit.regionMarkers[0].p2.pos_cal
            ):
                bad = True
                reason = "peak position outside of region"
            # TODO: check for NaNs in errors
            elif fit.fitter.peakModel.name == "theuerkauf":
                if (
                    peak.width.nominal_value <= 0.0
                    or peak.width.nominal_value > 5 * self.sigma_E
                ):
                    bad = True
                    reason = "width = %s" % peak.width
            elif fit.fitter.peakModel.name == "ee":
                # FIXME: what are frequent bad things that can happen here???
                pass
            if reason:
                text += "\n" + str(peak).strip("\n") + " (reason: " + reason + ")"
            else:
                text += "\n" + str(peak).strip("\n")
        if bad:
            return text


# FIXME or remove
#    def BGFit(self):
#        """
#        Do a background fit
#        """
#        # Background fit
#        sid = self.spectra.activeID
#        tSpec = ROOT.TSpectrum()
#
#        spec = self.spectra[self.spectra.activeID]

#        hist = spec.fHist.__class__(spec.fHist) # Copy hist here
#        # TODO: draw background
#        hbg = hist.ShowBackground(20, "goff")
#        print "hbg", hbg
#        bgspec = hdtv.spectrum.Spectrum(hbg, cal=spec.cal)
#
#        bspec = hdtv.spectrum.SpectrumCompound(spec.viewport, bgspec)
#        sid = self.spectra.Add(bgspec)
#        bgspec.SetColor(hdtv.color.ColorForID(sid))
#        print "BGSPec", sid


# plugin initialisation
import __main__

peakfinder = PeakFinder(__main__.spectra)
hdtv.cmdline.RegisterInteractive("peakfinder", peakfinder)


# wrapper function


def PeakSearch(args):
    try:
        if __main__.spectra.activeID not in __main__.spectra.visible:
            hdtv.ui.warning("Active spectrum is not visible, no action taken")
            return True
    except KeyError:
        raise hdtv.cmdline.HDTVCommandAbort("No active spectrum")
        return False

    sid = __main__.spectra.activeID

    if args.sigma is None:
        args.sigma = hdtv.options.Get("fit.peakfind.sigma")
    if args.threshold is None:
        args.threshold = hdtv.options.Get("fit.peakfind.threshold")
    if args.autofit is None:
        args.autofit = hdtv.options.Get("fit.peakfind.auto_fit")

    # TODO: Access session peakfinder
    peakfinder(
        sid, args.sigma, args.threshold, args.start, args.end, args.autofit, args.reject
    )


# Register configuration variables for "fit peakfind"
opt = hdtv.options.Option(default=2.5, parse=lambda x: float(x))
hdtv.options.RegisterOption("fit.peakfind.sigma", opt)
opt = hdtv.options.Option(default=0.05, parse=lambda x: float(x))
hdtv.options.RegisterOption("fit.peakfind.threshold", opt)
opt = hdtv.options.Option(default=False, parse=hdtv.options.parse_bool)
hdtv.options.RegisterOption("fit.peakfind.auto_fit", opt)

# Register command "fit peakfind"
prog = "fit peakfind"
description = "Search for peaks in active spectrum in given range"
parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
parser.add_argument(
    "-s", "--sigma", type=float, action="store", default=None, help="FWHM of peaks"
)
parser.add_argument(
    "-t",
    "--threshold",
    type=float,
    action="store",
    default=None,
    help="Threshold of peaks to accept in fraction of the amplitude of highest peak (0. < threshold < 1.)",
)
parser.add_argument(
    "-a",
    "--autofit",
    action="store_true",
    default=None,
    help="automatically fit found peaks",
)
parser.add_argument(
    "-r",
    "--reject",
    action="store_true",
    default=False,
    help="reject fits with unreasonable values",
)
parser.add_argument("start", nargs="?", type=float, default=None, help="start of range")
parser.add_argument("end", nargs="?", type=float, default=None, help="end of range")
hdtv.cmdline.AddCommand(prog, PeakSearch, level=4, parser=parser, fileargs=False)



# -*- coding: utf-8 -*-

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from .efficiency import _Efficiency
from ROOT import TF1
import math


class WunderEff(_Efficiency):
    """
    'Wunder' efficiency formula.

    eff(E) = (a*E + b/E) * exp(c*E + d/E)
    """

    def __init__(self, pars=list(), norm=True):

        self.name = "Wunder"
        self.id = self.name + "_" + hex(id(self))
        #                        ( N  * ( a *E +  b /E) * exp( c *E +  d /E))
        # [0] is fixed
        self.TF1 = TF1(self.id, "([0] * ([1]*x + [2]/x) * exp([3]*x + [4]/x))", 0, 0)

        _Efficiency.__init__(self, num_pars=5, pars=pars, norm=norm)

        # List of derivatives
        self._dEff_dP = [None, None, None, None, None]
        self._dEff_dP[0] = (
            lambda E, fPars: self.norm
            * (fPars[1] * E + fPars[2] / E)
            * math.exp(fPars[3] * E + fPars[4] / E)
        )  # dEff/dN
        self._dEff_dP[0] = (
            lambda E, fPars: self.norm
            * fPars[0]
            * E
            * math.exp(fPars[3] * E + fPars[4] / E)
        )  # dEff/da
        self._dEff_dP[1] = (
            lambda E, fPars: self.norm
            * fPars[0]
            * 1.0
            / E
            * math.exp(fPars[3] * E + fPars[4] / E)
        )  # dEff/db
        self._dEff_dP[2] = (
            lambda E, fPars: self.norm
            * fPars[0]
            * (fPars[1] * E + fPars[2] / E)
            * E
            * math.exp(fPars[3] * E + fPars[4] / E)
        )  # dEff/dc
        self._dEff_dP[3] = (
            lambda E, fPars: self.norm
            * fPars[0]
            * (fPars[1] * E + fPars[1] / E)
            * 1.0
            / E
            * math.exp(fPars[3] * E + fPars[4] / E)
        )  # dEff/dd
